from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import contextlib
import hashlib
import io
import logging
import os
import shutil

import chromadb
import chromadb.errors
from chromadb.api import ClientAPI
from chromadb.api.types import OneOrMany
from chromadb.config import Settings

from versionhq._utils.logger import Logger
from versionhq._utils.vars import KNOWLEDGE_DIRECTORY
from versionhq.storage.utils import fetch_db_storage_path
from versionhq.knowledge.embedding import EmbeddingConfigurator


@contextlib.contextmanager
def suppress_logging(
    logger_name="chromadb.segment.impl.vector.local_persistent_hnsw",
    level=logging.ERROR,
):
    logger = logging.getLogger(logger_name)
    original_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    with (
        contextlib.redirect_stdout(io.StringIO()),
        contextlib.redirect_stderr(io.StringIO()),
        contextlib.suppress(UserWarning),
    ):
        yield
    logger.setLevel(original_level)



class BaseKnowledgeStorage(ABC):
    """
    Abstract base class for knowledge storage implementations.
    """

    @abstractmethod
    def search(self, query: List[str], limit: int = 3, filter: Optional[dict] = None, score_threshold: float = 0.35) -> List[Dict[str, Any]]:
        """Search for documents in the knowledge base."""
        pass

    @abstractmethod
    def save(self, documents: List[str], metadata: Dict[str, Any] | List[Dict[str, Any]]) -> None:
        """Save documents to the knowledge base."""
        pass

    @abstractmethod
    def reset(self) -> None:
        """Reset the knowledge base."""
        pass



class KnowledgeStorage(BaseKnowledgeStorage):
    """
    Extends Storage to handle embeddings for memory entries, improving search efficiency.
    """

    collection: Optional[chromadb.Collection] = None
    collection_name: Optional[str] = "knowledge"
    app: Optional[ClientAPI] = None

    def __init__(self, embedder_config: Optional[Dict[str, Any]] = None, collection_name: Optional[str] = None):
        self.collection_name = collection_name
        self._set_embedder_config(embedder_config)

    def search(self, query: List[str], limit: int = 3, filter: Optional[dict] = None, score_threshold: float = 0.35) -> List[Dict[str, Any]]:
        with suppress_logging():
            if self.collection:
                fetched = self.collection.query(query_texts=query, n_results=limit, where=filter)
                results = []
                for i in range(len(fetched["ids"][0])):
                    result = {
                        "id": fetched["ids"][0][i],
                        "metadata": fetched["metadatas"][0][i],
                        "context": fetched["documents"][0][i],
                        "score": fetched["distances"][0][i],
                    }
                    if result["score"] >= score_threshold:
                        results.append(result)
                return results
            else:
                raise Exception("Collection not initialized")


    def initialize_knowledge_storage(self):
        base_path = os.path.join(fetch_db_storage_path(), "knowledge")
        chroma_client = chromadb.PersistentClient(path=base_path, settings=Settings(allow_reset=True))
        self.app = chroma_client

        try:
            collection_name = f"knowledge_{self.collection_name}" if self.collection_name else "knowledge"
            if self.app:
                self.collection = self.app.get_or_create_collection(name=collection_name, embedding_function=self.embedder_config)
            else:
                raise Exception("Vector Database Client not initialized")
        except Exception:
            raise Exception("Failed to create or get collection")


    def reset(self):
        base_path = os.path.join(fetch_db_storage_path(), KNOWLEDGE_DIRECTORY)
        if not self.app:
            self.app = chromadb.PersistentClient(path=base_path, settings=Settings(allow_reset=True))
        self.app.reset()
        shutil.rmtree(base_path)
        self.app = None
        self.collection = None


    def save(self, documents: List[str], metadata: Optional[Dict[str, Any] | List[Dict[str, Any]]] = None) -> None:
        if not self.collection:
            raise Exception("Collection not initialized")

        try:
            unique_docs = {}
            for i, doc in enumerate(documents):
                doc_id = hashlib.sha256(doc.encode("utf-8")).hexdigest()
                doc_metadata = None
                if metadata is not None:
                    if isinstance(metadata, list):
                        doc_metadata = metadata[i]
                    else:
                        doc_metadata = metadata
                unique_docs[doc_id] = (doc, doc_metadata)

            filtered_docs = []
            filtered_metadata = []
            filtered_ids = []

            for doc_id, (doc, meta) in unique_docs.items():
                filtered_docs.append(doc)
                filtered_metadata.append(meta)
                filtered_ids.append(doc_id)

            final_metadata: Optional[OneOrMany[chromadb.Metadata]] = (
                None if all(m is None for m in filtered_metadata) else filtered_metadata
            )
            self.collection.upsert(documents=filtered_docs, metadatas=final_metadata, ids=filtered_ids)

        except chromadb.errors.InvalidDimensionException as e:
            Logger(verbose=True).log(
                level="error",
                message="Embedding dimension mismatch. This usually happens when mixing different embedding models.",
                color="red",
            )
            raise ValueError("Embedding dimension mismatch. Make sure you're using the same embedding model across all operations with this collection.") from e

        except Exception as e:
            Logger(verbose=True).log(level="error", message=f"Failed to upsert documents: {str(e)}", color="red")
            raise


    def _create_default_embedding_function(self) -> Any:
        from chromadb.utils.embedding_functions.openai_embedding_function import (
            OpenAIEmbeddingFunction,
        )

        return OpenAIEmbeddingFunction(
            api_key=os.getenv("OPENAI_API_KEY"), model_name="text-embedding-3-small"
        )


    def _set_embedder_config(self, embedder_config: Optional[Dict[str, Any]] = None) -> None:
        """
        Set the embedding configuration for the knowledge storage.
        """
        self.embedder_config = EmbeddingConfigurator().configure_embedder(embedder_config) if embedder_config else self._create_default_embedding_function()



from enum import Enum
from typing import Type

JSON_URL = "https://raw.githubusercontent.com/BerriAI/litellm/main/model_prices_and_context_window.json"


"""
List of models available on the framework.
Model names align with the LiteLLM's key names defined in the JSON URL.
Provider names align with the custom provider or model provider names.
-> model_key = custom_provider_name/model_name

Option
litellm.pick_cheapest_chat_models_from_llm_provider(custom_llm_provider: str, n=1)
"""

MODELS = {
    "openai": [
        # "gpt-3.5-turbo",
        "gpt-4",
        "gpt-4o",
        "gpt-4o-mini",
        "o1-mini",
        "o1-preview",
    ],
    "gemini": [
        "gemini/gemini-1.5-flash",
        "gemini/gemini-1.5-pro",
        "gemini/gemini-2.0-flash-exp",
        # "gemini/gemini-gemma-2-9b-it",
        # "gemini/gemini-gemma-2-27b-it",
    ],
    # "vetrex_ai": [
    # ],
    "anthropic": [
        "claude-3-5-sonnet-20241022",
        "claude-3-5-sonnet-20240620",
        "claude-3-sonnet-20240229",
        "claude-3-opus-20240229",
        "claude-3-haiku-20240307",
    ],
    # "ollama": [
    #   "ollama/llama3.1",
    #   "ollama/mixtral",
    # ],
    # "watson": [
    #     "watsonx/meta-llama/llama-3-1-70b-instruct",
    #     "watsonx/meta-llama/llama-3-1-8b-instruct",
    #     "watsonx/meta-llama/llama-3-2-11b-vision-instruct",
    #     "watsonx/meta-llama/llama-3-2-1b-instruct",
    #     "watsonx/meta-llama/llama-3-2-90b-vision-instruct",
    #     "watsonx/meta-llama/llama-3-405b-instruct",
    #     "watsonx/mistral/mistral-large",
    #     "watsonx/ibm/granite-3-8b-instruct",
    # ],
    # "bedrock": [
    #     "bedrock/anthropic.claude-3-5-sonnet-20240620-v1:0",
    #     "bedrock/anthropic.claude-3-sonnet-20240229-v1:0",
    #     "bedrock/anthropic.claude-3-haiku-20240307-v1:0",
    #     "bedrock/anthropic.claude-3-opus-20240229-v1:0",
    #     "bedrock/anthropic.claude-v2:1",
    #     "bedrock/anthropic.claude-v2",
    #     "bedrock/anthropic.claude-instant-v1",
    #     "bedrock/meta.llama3-1-405b-instruct-v1:0",
    #     "bedrock/meta.llama3-1-70b-instruct-v1:0",
    #     "bedrock/meta.llama3-1-8b-instruct-v1:0",
    #     "bedrock/meta.llama3-70b-instruct-v1:0",
    #     "bedrock/meta.llama3-8b-instruct-v1:0",
    #     "bedrock/amazon.titan-text-lite-v1",
    #     "bedrock/amazon.titan-text-express-v1",
    #     "bedrock/cohere.command-text-v14",
    #     "bedrock/ai21.j2-mid-v1",
    #     "bedrock/ai21.j2-ultra-v1",
    #     "bedrock/ai21.jamba-instruct-v1:0",
    #     "bedrock/meta.llama2-13b-chat-v1",
    #     "bedrock/meta.llama2-70b-chat-v1",
    #     "bedrock/mistral.mistral-7b-instruct-v0:2",
    #     "bedrock/mistral.mixtral-8x7b-instruct-v0:1",
    # ],
}


PROVIDERS = [
    "openai",
    "anthropic",
    "gemini",
    "ollama",
    "watson",
    "bedrock",
    "azure",
    "cerebras",
    "llama",
]


"""
Max input token size by the model.
"""
LLM_CONTEXT_WINDOW_SIZES = {
    "gpt-3.5-turbo": 8192,
    "gpt-4": 8192,
    "gpt-4o": 128000,
    "gpt-4o-mini": 128000,
    "gpt-4-turbo": 128000,
    "o1-preview": 128000,
    "o1-mini": 128000,

    "gemini/gemini-1.5-flash": 1048576,
    "gemini/gemini-1.5-pro": 2097152,
    "gemini/gemini-2.0-flash-exp": 1048576,
    "gemini/gemini-gemma-2-9b-it": 8192,
    "gemini/gemini-gemma-2-27b-it": 8192,

    "claude-3-5-sonnet-20241022": 200000,
    "claude-3-5-sonnet-20240620": 200000,
    "claude-3-sonnet-20240229": 200000,
    "claude-3-opus-20240229": 200000,
    "claude-3-haiku-20240307": 200000,

    "deepseek-chat": 128000,
    "gemma2-9b-it": 8192,
    "gemma-7b-it": 8192,
    "llama3-groq-70b-8192-tool-use-preview": 8192,
    "llama3-groq-8b-8192-tool-use-preview": 8192,
    "llama-3.1-70b-versatile": 131072,
    "llama-3.1-8b-instant": 131072,
    "llama-3.2-1b-preview": 8192,
    "llama-3.2-3b-preview": 8192,
    "llama-3.2-11b-text-preview": 8192,
    "llama-3.2-90b-text-preview": 8192,
    "llama3-70b-8192": 8192,
    "llama3-8b-8192": 8192,
    "mixtral-8x7b-32768": 32768,
    "claude-3-5-sonnet-2024102": 200000,
}


LLM_API_KEY_NAMES = {
    "openai":  "OPENAI_API_KEY",
    "anthropic": "ANTHROPIC_API_KEY",
    "gemini": "GEMINI_API_KEY",
}

LLM_BASE_URL_KEY_NAMES = {
    "openai":  "OPENAI_API_BASE",
     "gemini": "GEMINI_API_BASE",
    "anthropic": "ANTHROPIC_API_BASE",
}

LLM_VARS = {
    "openai": [
        {
            "prompt": "Enter your OPENAI API key (press Enter to skip)",
            "key_name": "OPENAI_API_KEY",
        }
    ],
    "anthropic": [
        {
            "prompt": "Enter your ANTHROPIC API key (press Enter to skip)",
            "key_name": "ANTHROPIC_API_KEY",
        }
    ],
    "gemini": [
        {
            "prompt": "Enter your GEMINI API key (press Enter to skip)",
            "key_name": "GEMINI_API_KEY",
        }
    ],
    "watson": [
        {
            "prompt": "Enter your WATSONX URL (press Enter to skip)",
            "key_name": "WATSONX_URL",
        },
        {
            "prompt": "Enter your WATSONX API Key (press Enter to skip)",
            "key_name": "WATSONX_APIKEY",
        },
        {
            "prompt": "Enter your WATSONX Project Id (press Enter to skip)",
            "key_name": "WATSONX_PROJECT_ID",
        },
    ],
    "ollama": [
        {
            "default": True,
            "API_BASE": "http://localhost:11434",
        }
    ],
    "bedrock": [
        {
            "prompt": "Enter your AWS Access Key ID (press Enter to skip)",
            "key_name": "AWS_ACCESS_KEY_ID",
        },
        {
            "prompt": "Enter your AWS Secret Access Key (press Enter to skip)",
            "key_name": "AWS_SECRET_ACCESS_KEY",
        },
        {
            "prompt": "Enter your AWS Region Name (press Enter to skip)",
            "key_name": "AWS_REGION_NAME",
        },
    ],
    "azure": [
        {
            "prompt": "Enter your Azure deployment name (must start with 'azure/')",
            "key_name": "model",
        },
        {
            "prompt": "Enter your AZURE API key (press Enter to skip)",
            "key_name": "AZURE_API_KEY",
        },
        {
            "prompt": "Enter your AZURE API base URL (press Enter to skip)",
            "key_name": "AZURE_API_BASE",
        },
        {
            "prompt": "Enter your AZURE API version (press Enter to skip)",
            "key_name": "AZURE_API_VERSION",
        },
    ],
    "cerebras": [
        {
            "prompt": "Enter your Cerebras model name (must start with 'cerebras/')",
            "key_name": "model",
        },
        {
            "prompt": "Enter your Cerebras API version (press Enter to skip)",
            "key_name": "CEREBRAS_API_KEY",
        },
    ],
}



"""
Params for litellm.completion() func. Address common/unique params to each provider.
"""

PARAMS = {
    "litellm": [
        "api_base",
        "api_version,"
        "num_retries",
        "context_window_fallback_dict",
        "fallbacks",
        "metadata",
    ],
    "common": [
        "model",
        "messages",
        "temperature",
        "top_p",
        "max_tokens",
        "stream",
        "tools",
        "tool_choice",
        "response_format",
        "n",
        "stop",
        "base_url",
        "api_key",
    ],
    "openai": [
        "timeout",
        # "temperature",
        # "top_p",
        # "n",
        # "stream",
        "stream_options",
        # "stop",
        "max_compl,etion_tokens",
        # "max_tokens",
        "modalities",
        "prediction",
        "audio",
        "presence_penalty",
        "frequency_penalty",
        "logit_bias",
        "user",
        # "response_format",
        "seed",
        # "tools",
        # "tool_choice",
        "logprobs",
        "top_logprobs",
        "parallel_tool_calls",
        "extra_headers",
        "model_list"
    ],
    "gemini": [
        "topK",
    ]
}


class SchemaType:
    """
    A class to store/convert a LLM-valid schema type from the Python Type object.
    https://swagger.io/docs/specification/v3_0/data-models/data-types/
    https://cloud.google.com/vertex-ai/docs/reference/rest/v1/Schema#Type
    """

    def __init__(self, type: Type):
        self.type = type

    def convert(self) -> str:
        if self.type is None:
            return "string"

        if self.type is int:
            return "integer"
        elif self.type is float:
            return "number"
        elif self.type is str:
            return "string"
        elif self.type is dict:
            return "object"
        elif self.type is list:
            return "array"
        elif self.type is bool:
            return "boolean"
        else:
            return "string"



import logging
import json
import os
import sys
import threading
import warnings
import litellm
from litellm import JSONSchemaValidationError
from abc import ABC
from dotenv import load_dotenv
from litellm import get_supported_openai_params
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Type
from typing_extensions import Self

from pydantic import UUID4, BaseModel, Field, PrivateAttr, field_validator, model_validator, create_model, InstanceOf, ConfigDict
from pydantic_core import PydanticCustomError

from openai import OpenAI

from versionhq.llm.llm_vars import LLM_CONTEXT_WINDOW_SIZES, LLM_API_KEY_NAMES, LLM_BASE_URL_KEY_NAMES, MODELS, PARAMS, SchemaType
from versionhq.task import TaskOutputFormat
from versionhq.task.model import ResponseField, Task
from versionhq.tool.model import Tool, ToolSet
from versionhq._utils.logger import Logger


load_dotenv(override=True)
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")
LITELLM_API_BASE = os.environ.get("LITELLM_API_BASE")
DEFAULT_CONTEXT_WINDOW_SIZE = int(8192 * 0.75)
DEFAULT_MODEL_NAME = os.environ.get("DEFAULT_MODEL_NAME")

proxy_openai_client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY"), organization="versionhq", base_url=LITELLM_API_BASE)
openai_client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY"))


class FilteredStream:
    def __init__(self, original_stream):
        self._original_stream = original_stream
        self._lock = threading.Lock()

    def write(self, s) -> int:
        with self._lock:
            if (
                "Give Feedback / Get Help: https://github.com/BerriAI/litellm/issues/new"
                in s
                or "LiteLLM.Info: If you need to debug this error, use `os.environ['LITELLM_LOG'] = 'DEBUG'`"
                in s
            ):
                return 0
            return self._original_stream.write(s)

    def flush(self):
        with self._lock:
            return self._original_stream.flush()


@contextmanager
def suppress_warnings():
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = FilteredStream(old_stdout)
        sys.stderr = FilteredStream(old_stderr)

        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


class LLM(BaseModel):
    """
    An LLM class to store params except for response formats which will be given in the task handling process.
    Use LiteLLM to connect with the model of choice.
    Some optional params are passed by the agent, else follow the default settings of the model provider.
    Ref. https://docs.litellm.ai/docs/completion/input
    """

    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=True))
    _init_model_name: str = PrivateAttr(default=None)
    _tokens: int = PrivateAttr(default=0) # accumulate total tokens used for the call
    model_config = ConfigDict(extra="allow")

    model: str = Field(default=DEFAULT_MODEL_NAME)
    provider: Optional[str] = Field(default=None, description="model provider or custom model provider")
    base_url: Optional[str] = Field(default=None, description="api base of the model provider")
    api_key: Optional[str] = Field(default=None, description="api key of the model provider")

    # optional params
    timeout: Optional[float | int] = Field(default=None)
    max_tokens: Optional[int] = Field(default=None)
    max_completion_tokens: Optional[int] = Field(default=None)
    context_window_size: Optional[int] = Field(default=DEFAULT_CONTEXT_WINDOW_SIZE)
    callbacks: List[Any] = Field(default_factory=list)
    temperature: Optional[float] = Field(default=None)
    top_p: Optional[float] = Field(default=None)
    n: Optional[int] = Field(default=None)
    stop: Optional[str | List[str]] = Field(default=None)
    presence_penalty: Optional[float] = Field(default=None)
    frequency_penalty: Optional[float] = Field(default=None)
    logit_bias: Optional[Dict[int, float]] = Field(default=None)
    seed: Optional[int] = Field(default=None)
    logprobs: Optional[bool] = Field(default=None)
    top_logprobs: Optional[int] = Field(default=None)
    response_format: Optional[Any] = Field(default=None)
    tools: Optional[List[Dict[str, Any]]] = Field(default_factory=list, description="store a list of tool properties")

    # LiteLLM specific fields
    api_base: Optional[str] = Field(default=None, description="litellm specific field - api base of the model provider")
    api_version: Optional[str] = Field(default=None)
    num_retries: Optional[int] = Field(default=2)
    context_window_fallback_dict: Optional[Dict[str, Any]] = Field(default=None, description="A mapping of model to use if call fails due to context window error")
    fallbacks: Optional[List[Any]]= Field(default=None, description="A list of model names + params to be used, in case the initial call fails")
    metadata: Optional[Dict[str, Any]] = Field(default=None)

    litellm.drop_params = True
    litellm.set_verbose = True
    os.environ['LITELLM_LOG'] = 'DEBUG'

    @model_validator(mode="after")
    def validate_base_params(self) -> Self:
        """
        1) Set up a valid model name with the provider name using the MODEL list.
        * Assign a default model and provider based on the given information when no model key is found in the MODEL list.

        2) Set up other base parameters for the model and LiteLLM.
        """

        if self.model is None:
            self._logger.log(level="error", message="Model name is missing.", color="red")
            raise PydanticCustomError("model_missing", "The model name must be provided.", {})


        self._init_model_name = self.model
        self.model = None
        self._tokens = 0

        if self.provider and MODELS.get(self.provider):
            provider_model_list = MODELS.get(self.provider)
            for item in provider_model_list:
                if self.model is None:
                    if item == self._init_model_name:
                        self.model = item
                    elif self._init_model_name in item and self.model is None:
                        self.model = item
                    else:
                        temp_model = provider_model_list[0]
                        self._logger.log(level="warning", message=f"The provided model: {self._init_model_name} is not in the list. We'll assign a model: {temp_model} from the selected model provider: {self.provider}.", color="yellow")
                        self.model = temp_model

        else:
            for k, v in MODELS.items():
                for item in v:
                    if self.model is None:
                        if self._init_model_name == item:
                            self.model = item
                            self.provider = k

                        elif self.model is None and self._init_model_name in item:
                            self.model = item
                            self.provider = k

            if self.model is None:
                self._logger.log(level="warning", message=f"The provided model \'{self.model}\' is not in the list. We'll assign a default model.", color="yellow")
                self.model = DEFAULT_MODEL_NAME
                self.provider = "openai"


        if self.callbacks:
            self._set_callbacks(self.callbacks)

        self.context_window_size = self._get_context_window_size()

        api_key_name = self.provider.upper() + "_API_KEY" if self.provider else None
        if api_key_name:
            self.api_key = os.environ.get(api_key_name, None)

        base_url_key_name = self.provider.upper() + "_API_BASE" if self.provider else None
        if base_url_key_name:
            self.base_url = os.environ.get(base_url_key_name)
            self.api_base = self.base_url

        return self


    def _create_valid_params(self, config: Dict[str, Any], provider: str = None) -> Dict[str, Any]:
        params = dict()
        valid_keys = list()

        if not provider:
            valid_keys = PARAMS.get("litellm") + PARAMS.get("common") + PARAMS.get(self.provider) if self.provider else PARAMS.get("litellm") + PARAMS.get("common")
        else:
            valid_keys = PARAMS.get("common") + PARAMS.get(self.provider)

        for item in valid_keys:
            if hasattr(self, item) and getattr(self, item):
                params[item] = getattr(self, item)
            elif item in config:
                params[item] = config[item]

        return params


    def call(
        self,
        messages: List[Dict[str, str]],
        response_format: Optional[Dict[str, Any]] = None,
        tools: Optional[List[Tool | ToolSet | Any ]] = None,
        config: Optional[Dict[str, Any]] = {}, # any other conditions to pass on to the model.
        tool_res_as_final: bool = False
    ) -> str:
        """
        Execute LLM based on the agent's params and model params.
        """

        litellm.drop_params = True

        with suppress_warnings():
            if len(self.callbacks) > 0:
                self._set_callbacks(self.callbacks) # passed by agent

            try:
                provider = self.provider if self.provider else "openai"
                self.response_format = { "type": "json_object" } if tool_res_as_final == True else response_format

                if not tools:
                    params = self._create_valid_params(config=config)
                    res = litellm.completion(messages=messages, stream=False, **params)
                    self._tokens += int(res["usage"]["total_tokens"])
                    return res["choices"][0]["message"]["content"]

                else:
                    self.tools = [item.tool.properties if isinstance(item, ToolSet) else item.properties for item in tools]

                    if provider == "openai":
                        params = self._create_valid_params(config=config, provider=provider)
                        res = openai_client.chat.completions.create(messages=messages, model=self.model, tools=self.tools)
                        tool_calls = res.choices[0].message.tool_calls
                        tool_res = ""

                        for item in tool_calls:
                            func_name = item.function.name
                            func_args = item.function.arguments

                            if not isinstance(func_args, dict):
                                try:
                                    func_args = json.loads(json.dumps(eval(str(func_args))))
                                except:
                                    pass

                            for tool in tools:
                                if isinstance(tool, ToolSet) and (tool.tool.name == func_name or tool.tool.func.__name__ == func_name or func_name == "random_func"):
                                    tool_instance = tool.tool
                                    args = tool.kwargs
                                    tool_res_to_add = tool_instance.run(params=args)

                                    if tool_res_as_final:
                                        tool_res += str(tool_res_to_add)
                                    else:
                                        messages.append(res.choices[0].message)
                                        messages.append({ "role": "tool", "tool_call_id": item.id, "content": str(tool_res_to_add) })

                                else:
                                    try:
                                        tool_res_to_add = tool.run(params=func_args)
                                        if tool_res_as_final:
                                            tool_res += str(tool_res_to_add)
                                        else:
                                            messages.append(res.choices[0].message)
                                            messages.append({ "role": "tool", "tool_call_id": item.id, "content": str(tool_res_to_add) })
                                    except:
                                        pass

                        if tool_res_as_final:
                            return tool_res
                        else:
                            res = openai_client.chat.completions.create(messages=messages, model=self.model, tools=self.tools)
                            self._tokens += int(res["usage"]["total_tokens"])
                            return res.choices[0].message.content

            except JSONSchemaValidationError as e:
                self._logger.log(level="error", message="Raw Response: {}".format(e.raw_response), color="red")
                raise e

            except Exception as e:
                self._logger.log(level="error", message=f"{self.model} failed to execute: {str(e)}", color="red")
                if "litellm.RateLimitError" in str(e):
                    raise e


    def _supports_function_calling(self) -> bool:
        try:
            params = get_supported_openai_params(model=self.model)
            return "response_format" in params
        except Exception as e:
            self._logger.log(level="error", message=f"Failed to get supported params: {str(e)}", color="red")
            return False


    def _supports_stop_words(self) -> bool:
        try:
            params = get_supported_openai_params(model=self.model)
            return "stop" in params
        except Exception as e:
            self._logger.log(level="error", message=f"Failed to get supported params: {str(e)}", color="red")
            return False


    def _get_context_window_size(self) -> int:
        """
        Only use 75% of the context window size to avoid cutting the message in the middle.
        """
        return int(LLM_CONTEXT_WINDOW_SIZES.get(self.model) * 0.75) if LLM_CONTEXT_WINDOW_SIZES.get(self.model) is not None else DEFAULT_CONTEXT_WINDOW_SIZE


    def _set_callbacks(self, callbacks: List[Any]):
        callback_types = [type(callback) for callback in callbacks]
        for callback in litellm.success_callback[:]:
            if type(callback) in callback_types:
                litellm.success_callback.remove(callback)

        for callback in litellm._async_success_callback[:]:
            if type(callback) in callback_types:
                litellm._async_success_callback.remove(callback)

        litellm.callbacks = callbacks



from typing import Any, Dict, List, Optional

from versionhq.storage.rag_storage import RAGStorage
from versionhq.storage.ltm_sqlite_storage import LTMSQLiteStorage


class Memory:
    """
    Base class for memory.
    """

    def __init__(self, storage: RAGStorage):
        self.storage = storage


    def save(self, value: Any, metadata: Optional[Dict[str, Any]] = None, agent: Optional[str] = None) -> None:
        metadata = metadata or {}

        if agent:
            metadata["agent"] = agent
        self.storage.save(value, metadata)


    def search(self, query: str, limit: int = 3, score_threshold: float = 0.35) -> List[Any]:
        return self.storage.search(query=query, limit=limit, score_threshold=score_threshold)



class ShortTermMemoryItem:
    def __init__(
        self,
        data: Any,
        agent: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.data = data
        self.agent = agent
        self.metadata = metadata if metadata is not None else {}


class ShortTermMemory(Memory):
    """
    A class for managing transient data related to immediate tasks and interactions.
    - Type: stm
    - Storage: Mem0Storage | RAGStorage
    """

    def __init__(self, agent = None, embedder_config: Dict[str, Any] = None, storage=None, path=None):
        if hasattr(agent, "memory_config") and agent.memory_config is not None:
            self.memory_provider = agent.memory_config.get("provider")
        else:
            self.memory_provider = None

        if self.memory_provider == "mem0":
            try:
                from versionhq.storage.mem0_storage import Mem0Storage
            except ImportError:
                raise ImportError("Mem0 is not installed. Please install it with `uv pip install mem0ai`.")

            storage = Mem0Storage(type="stm", agent=agent)
        else:
            storage = storage if storage else RAGStorage(type="stm", embedder_config=embedder_config, agents=[agent,], path=path)

        super().__init__(storage)


    def save(self, value: Any, metadata: Optional[Dict[str, Any]] = None, agent: Optional[str] = None) -> None:
        item = ShortTermMemoryItem(data=value, metadata=metadata, agent=agent)
        if self.memory_provider == "mem0":
            item.data = f"Remember the following insights from Agent run: {item.data}"

        super().save(value=item.data, metadata=item.metadata, agent=item.agent)


    def search(self, query: str, limit: int = 3, score_threshold: float = 0.35,):
        return self.storage.search(query=query, limit=limit, score_threshold=score_threshold)


    def reset(self) -> None:
        try:
            self.storage.reset()
        except Exception as e:
            raise Exception(f"An error occurred while resetting the short-term memory: {str(e)}")



class LongTermMemoryItem:
    def __init__(
        self,
        agent: str,
        task: str,
        datetime: str,
        quality: Optional[int | float] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.task = task
        self.agent = agent
        self.quality = quality
        self.datetime = datetime
        self.metadata = metadata if metadata is not None else {}



class LongTermMemory(Memory):
    """
    A class for managing cross runs data related to overall task executions.
    - Type: ltm
    - Storage: LTMSQLiteStorage
    """

    def __init__(self, storage=None, path=None):
        if not storage:
            storage = LTMSQLiteStorage(db_path=path) if path else LTMSQLiteStorage()

        super().__init__(storage)


    def save(self, item: LongTermMemoryItem) -> None:
        metadata = item.metadata
        metadata.update({ "agent": item.agent })
        self.storage.save(
            task_description=item.task,
            score=metadata["quality"],
            metadata=metadata,
            datetime=item.datetime,
        )


    def search(self, task: str, latest_n: int = 3) -> List[Dict[str, Any]]:
        return self.storage.load(task, latest_n)


    def reset(self) -> None:
        self.storage.reset()



class UserMemoryItem:
    def __init__(self, data: Any, user: str, metadata: Optional[Dict[str, Any]] = None):
        self.data = data
        self.user = user
        self.metadata = metadata if metadata is not None else {} # can be stored last purchased item, comm related to the user


class UserMemory(Memory):
    """
    UserMemory class for handling user memory storage and retrieval.
    - Type: user
    - Storage: Mem0Storage
    - Requirements: `user_id` in metadata
    """

    def __init__(self, agent=None, user_id=None):
        try:
            from versionhq.storage.mem0_storage import Mem0Storage
        except ImportError:
            raise ImportError("Mem0 is not installed. Please install it with `uv pip install mem0ai`.")

        if not user_id:
            raise ValueError("Need User Id to create UserMemory.")

        else:
            storage = Mem0Storage(type="user", agent=agent, user_id=user_id)
            super().__init__(storage)


    def save(self, value: str, metadata: Optional[Dict[str, Any]] = None, agent: Optional[str] = None) -> None:
        data = f"Remember the details about the user: {value}"
        super().save(value=data, metadata=metadata, agent=agent)


    def search(self, query: str, limit: int = 3, score_threshold: float = 0.35):
        results = self.storage.search(query=query, limit=limit, score_threshold=score_threshold)
        return results



import uuid
import warnings
import json
from enum import Enum
from dotenv import load_dotenv
from concurrent.futures import Future
from hashlib import md5
from typing import Any, Dict, List, TYPE_CHECKING, Callable, Optional, Tuple
from pydantic import UUID4, InstanceOf, Json, BaseModel, Field, PrivateAttr, field_validator, model_validator
from pydantic._internal._generate_schema import GenerateSchema
from pydantic_core import PydanticCustomError, core_schema

from versionhq.agent.model import Agent
from versionhq.task.model import Task, TaskOutput, ConditionalTask, TaskOutputFormat
from versionhq.task.formatter import create_raw_outputs
from versionhq.team.team_planner import TeamPlanner
from versionhq._utils.logger import Logger
from versionhq._utils.usage_metrics import UsageMetrics


initial_match_type = GenerateSchema.match_type


def match_type(self, obj):
    if getattr(obj, "__name__", None) == "datetime":
        return core_schema.datetime_schema()
    return initial_match_type(self, obj)


GenerateSchema.match_type = match_type

warnings.filterwarnings("ignore", category=SyntaxWarning, module="pysbd")
load_dotenv(override=True)

# agentops = None
# if os.environ.get("AGENTOPS_API_KEY"):
#     try:
#         import agentops  # type: ignore
#     except ImportError:
#         pass


class TaskHandlingProcess(str, Enum):
    """
    Class representing the different processes that can be used to tackle multiple tasks.
    """
    sequential = "sequential"
    hierarchical = "hierarchical"
    consensual = "consensual"


class TeamOutput(BaseModel):
    """
    Store outputs of the tasks handled by the team.
    `json_dict` and `raw` store overall output of tasks that handled by the team,
    while `task_output_list` stores each TaskOutput instance to the tasks handled by the team members.
    Note that `raw` and `json_dict` will be prioritized as TeamOutput to refer over `task_output_list`.
    """

    team_id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="store the team ID that generate the TeamOutput")
    raw: str = Field(default="", description="raw output of the team lead task handled by the team leader")
    pydantic: Optional[Any] = Field(default=None, description="`raw` converted to the abs. pydantic model")
    json_dict: Dict[str, Any] = Field(default=None, description="`raw` converted to dictionary")
    task_output_list: list[TaskOutput] = Field(default=list, description="store output of all the tasks that the team has executed")
    token_usage: UsageMetrics = Field(default=dict, description="processed token summary")

    def __str__(self):
        return (str(self.pydantic) if self.pydantic else str(self.json_dict) if self.json_dict else self.raw)


    def __getitem__(self, key):
        if self.pydantic and hasattr(self.pydantic, key):
            return getattr(self.pydantic, key)
        elif self.json_dict and key in self.json_dict:
            return self.json_dict[key]
        else:
            raise KeyError(f"Key '{key}' not found in the team output.")


    @property
    def json(self) -> Optional[str]:
        if self.tasks_output[-1].output_format != TaskOutputFormat.JSON:
            raise ValueError(
                "No JSON output found in the final task. Please make sure to set the output_json property in the final task in your team."
            )
        return json.dumps(self.json_dict)


    def to_dict(self) -> Dict[str, Any]:
        """
        Convert pydantic / raw output into dict and return the dict.
        When we only have `raw` output, return `{ output: raw }` to avoid an error
        """

        output_dict = {}
        if self.json_dict:
            output_dict.update(self.json_dict)
        elif self.pydantic:
            output_dict.update(self.pydantic.model_dump())
        else:
            output_dict.upate({ "output": self.raw })
        return output_dict


    def return_all_task_outputs(self) -> List[Dict[str, Any]]:
        res = [output.json_dict for output in self.task_output_list]
        return res


class TeamMember(BaseModel):
    agent: Agent | None = Field(default=None, description="store the agent to be a member")
    is_manager: bool = Field(default=False)
    task: Optional[Task] = Field(default=None)

    @property
    def is_idling(self):
        return bool(self.task is None)


class Team(BaseModel):
    """
    A collaborative team of agents that handles complex, multiple tasks.
    We define strategies for task executions and overall workflow.
    """

    __hash__ = object.__hash__
    _execution_span: Any = PrivateAttr()
    _logger: Logger = PrivateAttr()
    _inputs: Optional[Dict[str, Any]] = PrivateAttr(default=None)

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None)
    members: List[TeamMember] = Field(default_factory=list, description="store agents' uuids and bool if it is manager")

    # work as a team
    team_tasks: Optional[List[Task]] = Field(default_factory=list, description="optional tasks for the team")
    planning_llm: Optional[Any] = Field(default=None, description="llm to handle the planning of the team tasks (if any)")
    function_calling_llm: Optional[Any] = Field(default=None, description="llm to execute func after all agent execution (if any)")
    prompt_file: str = Field(default="", description="path to the prompt json file to be used by the team.")
    process: TaskHandlingProcess = Field(default=TaskHandlingProcess.sequential)

    # callbacks
    before_kickoff_callbacks: List[Callable[[Optional[Dict[str, Any]]], Optional[Dict[str, Any]]]] = Field(
        default_factory=list,
        description="list of callback functions to be executed before the team kickoff. i.e., adjust inputs"
    )
    after_kickoff_callbacks: List[Callable[[TeamOutput], TeamOutput]] = Field(
        default_factory=list,
        description="list of callback functions to be executed after the team kickoff. i.e., store the result in repo"
    )
    step_callback: Optional[Any] = Field(default=None, description="callback to be executed after each step for all agents execution")

    verbose: bool = Field(default=True)
    cache: bool = Field(default=True)
    memory: bool = Field(default=False, description="whether the team should use memory to store memories of its execution")
    execution_logs: List[Dict[str, Any]] = Field(default=[], description="list of execution logs for tasks")
    usage_metrics: Optional[UsageMetrics] = Field(default=None, description="usage metrics for all the llm executions")


    def __name__(self) -> str:
        return self.name if self.name is not None else self.id.__str__


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        """Prevent manual setting of the 'id' field by users."""
        if v:
            raise PydanticCustomError("may_not_set_field", "The 'id' field cannot be set by the user.", {})


    @model_validator(mode="after")
    def assess_tasks(self):
        """
        Validates if the model recognize all tasks that the team needs to handle.
        """

        if self.tasks:
            if all(task in self.tasks for task in self.team_tasks) == False:
                raise PydanticCustomError("task_validation_error", "`team_tasks` needs to be recognized in the task.", {})

            if len(self.tasks) != len(self.team_tasks) + len([member for member in self.members if member.task is not None]):
                raise PydanticCustomError("task_validation_error", "Some tasks are missing.", {})
        return self


    @model_validator(mode="after")
    def check_manager_llm(self):
        """
        Validates that the language model is set when using hierarchical process.
        """

        if self.process == TaskHandlingProcess.hierarchical:
            if self.managers is None:
                raise PydanticCustomError(
                    "missing_manager_llm_or_manager",
                    "Attribute `manager_llm` or `manager` is required when using hierarchical process.",
                    {},
                )

            if self.managers and (self.manager_tasks is None or self.team_tasks is None):
                raise PydanticCustomError("missing_manager_task", "manager needs to have at least one manager task or team task.", {})

        return self


    @model_validator(mode="after")
    def validate_tasks(self):
        """
        Sequential task processing without any team tasks require a task-agent pairing.
        """

        if self.process == TaskHandlingProcess.sequential and self.team_tasks is None:
            for member in self.members:
                if member.task is None:
                    raise PydanticCustomError("missing_agent_in_task", "Sequential process error: Agent is missing the task", {})
        return self

    @model_validator(mode="after")
    def validate_end_with_at_most_one_async_task(self):
        """
        Validates that the team completes max. one asynchronous task by counting tasks traversed backward
        """

        async_task_count = 0
        for task in reversed(self.tasks):
            if not task:
                break
            elif task.async_execution:
                async_task_count += 1
            else:
                break

        if async_task_count > 1:
            raise PydanticCustomError("async_task_count", "The team must end with max. one asynchronous task.", {})
        return self


    def _get_responsible_agent(self, task: Task) -> Agent:
        if task is None:
            return None
        else:
            res = [member.agent for member in self.members if member.task and member.task.id == task.id]
            return None if len(res) == 0 else res[0]


    def _handle_team_planning(self) -> None:
        """
        Form a team considering agents and tasks given, and update `self.members` field:
            1. Idling managers to take the team tasks.
            2. Idling members to take the remaining tasks starting from the team tasks to member tasks.
            3. Create agents to handle the rest tasks.
        """

        team_planner = TeamPlanner(tasks=self.tasks, planner_llm=self.planning_llm)
        idling_managers: List[TeamMember] = [member for member in self.members if member.is_idling and member.is_manager is True]
        idling_members: List[TeamMember] =  [member for member in self.members if member.is_idling and member.is_manager is False]
        unassigned_tasks: List[Task] = self.member_tasks_without_agent
        new_team_members: List[TeamMember] = []

        if self.team_tasks:
            candidates = idling_managers + idling_members
            if candidates:
                i = 0
                while i < len(candidates):
                    if self.team_tasks[i]:
                        candidates[i].task = self.team_tasks[i]
                        i += 1

                if len(self.team_tasks) > i:
                    for item in self.team_tasks[i:]:
                        if item not in unassigned_tasks:
                            unassigned_tasks = [item, ] + unassigned_tasks

            else:
                for item in self.team_tasks:
                    if item not in unassigned_tasks:
                        unassigned_tasks = [item, ] + unassigned_tasks

        if unassigned_tasks:
            new_team_members = team_planner._handle_assign_agents(unassigned_tasks=unassigned_tasks)

        if new_team_members:
            self.members += new_team_members


    # task execution
    def _process_async_tasks(
        self, futures: List[Tuple[Task, Future[TaskOutput], int]], was_replayed: bool = False
    ) -> List[TaskOutput]:
        """
        When we have `Future` tasks, updated task outputs and task execution logs accordingly.
        """

        task_outputs: List[TaskOutput] = []

        for future_task, future, task_index in futures:
            task_output = future.result()
            task_outputs.append(task_output)
            future_task._store_execution_log(task_index, was_replayed)

        return task_outputs


    def _create_team_output(self, task_outputs: List[TaskOutput], lead_task_output: TaskOutput = None) -> TeamOutput:
        """
        Take the output of the first task or the lead task output as the team output `raw` value.
        Note that `tasks` are already sorted by the importance.
        """

        if len(task_outputs) < 1:
            raise ValueError("Something went wrong. Kickoff should return only one task output.")

        final_task_output = lead_task_output if lead_task_output is not None else task_outputs[0]
        # final_string_output = final_task_output.raw
        # self._finish_execution(final_string_output)
        token_usage = self._calculate_usage_metrics()

        return TeamOutput(
            team_id=self.id,
            raw=final_task_output.raw,
            json_dict=final_task_output.json_dict,
            pydantic=final_task_output.pydantic,
            task_output_list=task_outputs,
            token_usage=token_usage,
        )


    def _calculate_usage_metrics(self) -> UsageMetrics:
        """
        Calculate and return the usage metrics that consumed by the team.
        """
        total_usage_metrics = UsageMetrics()

        for member in self.members:
            agent = member.agent
            if hasattr(agent, "_token_process"):
                token_sum = agent._token_process.get_summary()
                total_usage_metrics.add_usage_metrics(token_sum)

        if self.managers:
            for manager in self.managers:
                if hasattr(manager.agent, "_token_process"):
                    token_sum = manager.agent._token_process.get_summary()
                    total_usage_metrics.add_usage_metrics(token_sum)

        self.usage_metrics = total_usage_metrics
        return total_usage_metrics


    def _execute_tasks(self, tasks: List[Task], start_index: Optional[int] = 0, was_replayed: bool = False) -> TeamOutput:
        """
        Executes tasks sequentially and returns the final output in TeamOutput class.
        When we have a manager agent, we will start from executing manager agent's tasks.
        Priority:
        1. Team tasks > 2. Manager task > 3. Member tasks (in order of index)
        """

        task_outputs: List[TaskOutput] = []
        lead_task_output: TaskOutput = None
        futures: List[Tuple[Task, Future[TaskOutput], int]] = []
        last_sync_output: Optional[TaskOutput] = None

        for task_index, task in enumerate(tasks):
            if start_index is not None and task_index < start_index:
                if task.output:
                    if task.async_execution:
                        task_outputs.append(task.output)
                    else:
                        task_outputs = [task.output]
                        last_sync_output = task.output
                continue

            responsible_agent = self._get_responsible_agent(task)
            if responsible_agent is None:
                self._handle_team_planning()

            if isinstance(task, ConditionalTask):
                skipped_task_output = task._handle_conditional_task(task_outputs, futures, task_index, was_replayed)
                if skipped_task_output:
                    continue

            # self._log_task_start(task, responsible_agent)

            if task.async_execution:
                context = create_raw_outputs(tasks=[task, ], task_outputs=([last_sync_output,] if last_sync_output else []))
                future = task.execute_async(agent=responsible_agent, context=context)
                futures.append((task, future, task_index))
            else:
                context = create_raw_outputs(tasks=[task,], task_outputs=([last_sync_output,] if last_sync_output else [] ))
                task_output = task.execute_sync(agent=responsible_agent, context=context)
                if self.managers and responsible_agent in [manager.agent for manager in self.managers]:
                    lead_task_output = task_output

                task_outputs.append(task_output)
                # self._process_task_result(task, task_output)
                task._store_execution_log(task_index, was_replayed, self._inputs)


        if futures:
            task_outputs = self._process_async_tasks(futures, was_replayed)

        return self._create_team_output(task_outputs, lead_task_output)


    def kickoff(self, kwargs_before: Optional[Dict[str, str]] = None, kwargs_after: Optional[Dict[str, Any]] = None) -> TeamOutput:
        """
        Kickoff the team:
        0. Assign an agent to a task - using conditions (manager prioritizes team_tasks) and planning_llm.
        1. Address `before_kickoff_callbacks` if any.
        2. Handle team members' tasks in accordance with the process.
        3. Address `after_kickoff_callbacks` if any.
        """

        metrics: List[UsageMetrics] = []

        if self.team_tasks or self.member_tasks_without_agent:
            self._handle_team_planning()

        if kwargs_before is not None:
            for before_callback in self.before_kickoff_callbacks:
                before_callback(**kwargs_before)

        # self._execution_span = self._telemetry.team_execution_span(self, inputs)
        # self._task_output_handler.reset()
        # self._logging_color = "bold_purple"


        # i18n = I18N(prompt_file=self.prompt_file)

        for member in self.members:
            agent = member.agent
            agent.team = self

            if not agent.function_calling_llm and self.function_calling_llm:
                agent.function_calling_llm = self.function_calling_llm

            if self.step_callback:
                agent.callbacks.append(self.step_callback)

        if self.process is None:
            self.process = TaskHandlingProcess.sequential

        result = self._execute_tasks(self.tasks)

        for after_callback in self.after_kickoff_callbacks:
            result = after_callback(result, **kwargs_after)

        metrics += [member.agent._token_process.get_summary() for member in self.members]

        self.usage_metrics = UsageMetrics()
        for metric in metrics:
            self.usage_metrics.add_usage_metrics(metric)

        return result


    @property
    def key(self) -> str:
        source = [str(member.agent.id.__str__) for member in self.members] + [str(task.id.__str__) for task in self.tasks]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()


    @property
    def managers(self) -> List[TeamMember] | None:
        managers = [member for member in self.members if member.is_manager == True]
        return managers if len(managers) > 0 else None


    @property
    def manager_tasks(self) -> List[Task] | None:
        """
        Tasks (incl. team tasks) handled by managers in the team.
        """
        if self.managers:
            tasks = [manager.task for manager in self.managers if manager.task is not None]
            return tasks if len(tasks) > 0 else None

        return None


    @property
    def tasks(self):
        """
        Return all the tasks that the team needs to handle in order of priority:
        1. team tasks, -> assigned to the member
        2. manager_task,
        3. members' tasks
        """

        team_tasks = self.team_tasks
        manager_tasks = [member.task for member in self.members if member.is_manager == True and member.task is not None and member.task not in team_tasks]
        member_tasks = [member.task for member in self.members if member.is_manager == False and member.task is not None and member.task not in team_tasks]

        return team_tasks + manager_tasks + member_tasks


    @property
    def member_tasks_without_agent(self) -> List[Task] | None:
        if self.members:
            return [member.task for member in self.members if member.agent is None]

        return None



from unittest.mock import patch

import pytest

from versionhq.agent.model import Agent
from versionhq.task.model import Task
from versionhq.memory.model import ShortTermMemory, ShortTermMemoryItem, LongTermMemory, LongTermMemoryItem


@pytest.fixture
def short_term_memory():
    """Fixture to create a ShortTermMemory instance"""
    agent = Agent(
        role="Demo",
        goal="Search relevant data and provide results",
    )

    task = Task(
        description="Perform a search on specific topics.",
        expected_output="A list of relevant URLs based on the search query.",
        agent=agent,
    )
    return ShortTermMemory(agent=agent)


def test_save_and_search_stm(short_term_memory):
    memory = ShortTermMemoryItem(
        data="""test value test value test value test value test value test value
        test value test value test value test value test value test value
        test value test value test value test value test value test value""",
        agent="test_agent",
        metadata={ "task": "test_task" },
    )

    with patch.object(ShortTermMemory, "save") as mock_save:
        short_term_memory.save(
            value=memory.data,
            metadata=memory.metadata,
            agent=memory.agent,
        )

        mock_save.assert_called_once_with(
            value=memory.data,
            metadata=memory.metadata,
            agent=memory.agent,
        )

    expected_result = [
        {
            "context": memory.data,
            "metadata": {"agent": "test_agent"},
            "score": 0.95,
        }
    ]
    with patch.object(ShortTermMemory, "search", return_value=expected_result):
        find = short_term_memory.search("test value", score_threshold=0.01)[0]
        assert find["context"] == memory.data, "Data value mismatch."
        assert find["metadata"]["agent"] == "test_agent", "Agent value mismatch."




@pytest.fixture
def long_term_memory():
    """Fixture to create a LongTermMemory instance"""
    return LongTermMemory()


def test_save_and_search_ltm(long_term_memory):
    memory = LongTermMemoryItem(
        agent="test_agent",
        task="test_task",
        datetime="test_datetime",
        quality=0.5,
        metadata={"task": "test_task", "quality": 0.5},
    )
    long_term_memory.save(memory)
    find = long_term_memory.search("test_task", latest_n = 5)[0]
    assert find["score"] == 0.5
    assert find["datetime"] == "test_datetime"
    assert find["metadata"]["agent"] == "test_agent"
    assert find["metadata"]["quality"] == 0.5
    assert find["metadata"]["task"] == "test_task"



EVALUATE="""Assess the accuracy and quality of the following task output to the task described below. Score based on the criterion (0-1, 0=worst, 1=best) and suggest improvements. Vary scores; don't assign identical values. Store criteria in the "criteria" field.
Task: {task_description}
Task Output: {task_output}
Evaluation criteria: {eval_criteria}
"""



import csv
import json
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional
from pathlib import Path

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, field_validator

from versionhq.knowledge.storage import KnowledgeStorage
from versionhq._utils.vars import KNOWLEDGE_DIRECTORY
from versionhq._utils.logger import Logger


class BaseKnowledgeSource(BaseModel, ABC):
    """
    Abstract base class for knowledge sources: csv, json, excel, pdf, string, and docling.
    """

    chunk_size: int = 4000
    chunk_overlap: int = 200
    chunks: List[str] = Field(default_factory=list)
    chunk_embeddings: List[np.ndarray] = Field(default_factory=list)

    model_config = ConfigDict(arbitrary_types_allowed=True)
    storage: Optional[KnowledgeStorage] = Field(default=None)
    metadata: Dict[str, Any] = Field(default_factory=dict)  # Currently unused
    collection_name: Optional[str] = Field(default=None)

    @abstractmethod
    def validate_content(self) -> Any:
        """Load and preprocess content from the source."""
        pass

    @abstractmethod
    def add(self) -> None:
        """Process content, chunk it, compute embeddings, and save them."""
        pass

    def get_embeddings(self) -> List[np.ndarray]:
        """Return the list of embeddings for the chunks."""
        return self.chunk_embeddings

    def _chunk_text(self, text: str) -> List[str]:
        """
        Utility method to split text into chunks.
        """

        return [
            text[i : i + self.chunk_size]
            for i in range(0, len(text), self.chunk_size - self.chunk_overlap)
        ]

    def _save_documents(self):
        """
        Save the documents to the storage.
        This method should be called after the chunks and embeddings are generated.
        """
        if self.storage:
            self.storage.save(self.chunks)
        else:
            raise ValueError("No storage found to save documents.")



class StringKnowledgeSource(BaseKnowledgeSource):
    """
    A knowledge source that stores and queries plain text content using embeddings.
    """

    content: str = Field(...)
    collection_name: Optional[str] = Field(default=None)

    def model_post_init(self, _):
        """Post-initialization method to validate content."""
        self.validate_content()

    def validate_content(self):
        """Validate string content."""
        if not isinstance(self.content, str):
            raise ValueError("StringKnowledgeSource only accepts string content")

    def add(self) -> None:
        """
        Add string content to the knowledge source, chunk it, compute embeddings, and save them.
        """
        new_chunks = self._chunk_text(self.content)
        self.chunks.extend(new_chunks)
        self._save_documents()


    def _chunk_text(self, text: str) -> List[str]:
        """
        Utility method to split text into chunks.
        """
        return [text[i : i + self.chunk_size] for i in range(0, len(text), self.chunk_size - self.chunk_overlap)]



class BaseFileKnowledgeSource(BaseKnowledgeSource, ABC):
    """Base class for knowledge sources that load content from files."""

    _logger: Logger = Logger(verbose=True)
    file_paths: Optional[Path | List[Path] | str | List[str]] = Field(default_factory=list)
    content: Dict[Path, str] = Field(init=False, default_factory=dict)
    storage: Optional[KnowledgeStorage] = Field(default=None)
    safe_file_paths: List[Path] = Field(default_factory=list, description="store a list of `Path` objects from self.file_paths")


    @field_validator("file_paths", mode="before")
    def validate_file_path(cls, v, info):
        """
        Validate if at least one valid file path is provided.
        """
        if v is None and info.data.get("file_paths") is None:
            raise ValueError("Either file_path or file_paths must be provided")
        return v


    def model_post_init(self, _) -> None:
        """
        Post-initialization method to load content.
        """
        self.safe_file_paths = self._process_file_paths()
        self.validate_content()
        self.content = self.load_content()


    @abstractmethod
    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess file content. Should be overridden by subclasses.
        Assume that the file path is relative to the project root in the knowledge directory.
        """
        pass


    def validate_content(self):
        """
        Validate the given file paths.
        """
        for path in self.safe_file_paths:
            if not path.exists():
                self._logger.log(
                    "error",
                    f"File not found: {path}. Try adding sources to the knowledge directory. If it's inside the knowledge directory, use the relative path.",
                    color="red",
                )
                raise FileNotFoundError(f"File not found: {path}")
            if not path.is_file():
                self._logger.log("error", f"Path is not a file: {path}", color="red")


    def _save_documents(self):
        if self.storage:
            self.storage.save(self.chunks)
        else:
            raise ValueError("No storage found to save documents.")


    def convert_to_path(self, path: Path | str) -> Path:
        """
        Convert a path to a Path object.
        """
        return Path(KNOWLEDGE_DIRECTORY + "/" + path) if isinstance(path, str) else path


    def _process_file_paths(self) -> List[Path]:
        """
        Convert file_path to a list of Path objects.
        """

        if self.file_paths is None:
            raise ValueError("Your source must be provided with a file_paths: []")

        path_list: List[Path | str] = [self.file_paths] if isinstance(self.file_paths, (str, Path)) else list(self.file_paths) if isinstance(self.file_paths, list) else []

        if not path_list:
            raise ValueError(
                "file_path/file_paths must be a Path, str, or a list of these types"
            )

        return [self.convert_to_path(path) for path in path_list]



class TextFileKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source class that stores and queries text file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess text file content.
        """

        content = {}
        for path in self.safe_file_paths:
            path = self.convert_to_path(path)
            with open(path, "r", encoding="utf-8") as f:
                content[path] = f.read()
        return content


    def add(self) -> None:
        """
        Add text file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        for _, text in self.content.items():
            new_chunks = self._chunk_text(text)
            self.chunks.extend(new_chunks)
        self._save_documents()


    def _chunk_text(self, text: str) -> List[str]:
        """
        Utility method to split text into chunks.
        """
        return [text[i:i + self.chunk_size] for i in range(0, len(text), self.chunk_size - self.chunk_overlap)]



class PDFKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source class that stores and queries PDF file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess PDF file content.
        """
        pdfplumber = self._import_pdfplumber()
        content = {}
        for path in self.safe_file_paths:
            text = ""
            path = self.convert_to_path(path)
            with pdfplumber.open(path) as pdf:
                for page in pdf.pages:
                    page_text = page.extract_text()
                    if page_text:
                        text += page_text + "\n"
            content[path] = text
        return content


    def _import_pdfplumber(self):
        """
        Dynamically import pdfplumber.
        """
        try:
            import pdfplumber
            return pdfplumber
        except ImportError:
            raise ImportError("pdfplumber is not installed. Please install it with: pip install pdfplumber")


    def add(self) -> None:
        """
        Add PDF file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        for _, text in self.content.items():
            new_chunks = self._chunk_text(text)
            self.chunks.extend(new_chunks)
        self._save_documents()


    def _chunk_text(self, text: str) -> List[str]:
        """
        Utility method to split text into chunks.
        """
        return [text[i : i + self.chunk_size] for i in range(0, len(text), self.chunk_size - self.chunk_overlap)]



class CSVKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source class that stores and queries CSV file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess CSV file content.
        """
        content_dict = {}
        for file_path in self.safe_file_paths:
            with open(file_path, "r", encoding="utf-8") as csvfile:
                reader = csv.reader(csvfile)
                content = ""
                for row in reader:
                    content += " ".join(row) + "\n"
                content_dict[file_path] = content

        return content_dict


    def add(self) -> None:
        """
        Add CSV file content to the knowledge source, chunk it, compute embeddings,
        and save the embeddings.
        """
        content_str = str(self.content) if isinstance(self.content, dict) else self.content
        new_chunks = self._chunk_text(content_str)
        self.chunks.extend(new_chunks)
        self._save_documents()


    def _chunk_text(self, text: str) -> List[str]:
        """
        Utility method to split text into chunks.
        """
        return [text[i:i + self.chunk_size] for i in range(0, len(text), self.chunk_size - self.chunk_overlap)]



class JSONKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source class that stores and queries JSON file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess JSON file content.
        """
        content: Dict[Path, str] = {}
        for path in self.safe_file_paths:
            path = self.convert_to_path(path)
            with open(path, "r", encoding="utf-8") as json_file:
                data = json.load(json_file)
            content[path] = self._json_to_text(data)
        return content

    def _json_to_text(self, data: Any, level: int = 0) -> str:
        """
        Recursively convert JSON data to a text representation.
        """
        text = ""
        indent = "  " * level
        if isinstance(data, dict):
            for key, value in data.items():
                text += f"{indent}{key}: {self._json_to_text(value, level + 1)}\n"
        elif isinstance(data, list):
            for item in data:
                text += f"{indent}- {self._json_to_text(item, level + 1)}\n"
        else:
            text += f"{str(data)}"
        return text


    def add(self) -> None:
        """
        Add JSON file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        content_str = str(self.content) if isinstance(self.content, dict) else self.content
        new_chunks = self._chunk_text(content_str)
        self.chunks.extend(new_chunks)
        self._save_documents()


    def _chunk_text(self, text: str) -> List[str]:
        """
        Utility method to split text into chunks.
        """
        return [text[i:i + self.chunk_size] for i in range(0, len(text), self.chunk_size - self.chunk_overlap)]



class ExcelKnowledgeSource(BaseFileKnowledgeSource):
    """
    A knowledge source that stores and queries Excel file content using embeddings.
    """

    def load_content(self) -> Dict[Path, str]:
        """
        Load and preprocess Excel file content.
        """

        pd = self._import_dependencies()
        content_dict = {}
        for file_path in self.safe_file_paths:
            file_path = self.convert_to_path(file_path)
            df = pd.read_excel(file_path)
            content = df.to_csv(index=False)
            content_dict[file_path] = content
        return content_dict

    def _import_dependencies(self):
        """
        Dynamically import dependencies.
        """
        try:
            import pandas as pd
            return pd
        except ImportError as e:
            missing_package = str(e).split()[-1]
            raise ImportError(
                f"{missing_package} is not installed. Please install it with: pip install {missing_package}"
            )

    def add(self) -> None:
        """
        Add Excel file content to the knowledge source, chunk it, compute embeddings, and save the embeddings.
        """
        content_str = "\n".join(str(value) for value in self.content.values()) if isinstance(self.content, dict) else str(self.content)
        new_chunks = self._chunk_text(content_str)
        self.chunks.extend(new_chunks)
        self._save_documents()


    def _chunk_text(self, text: str) -> List[str]:
        """
        Utility method to split text into chunks.
        """
        return [text[i:i + self.chunk_size] for i in range(0, len(text), self.chunk_size - self.chunk_overlap)]



from versionhq.agent.model import Agent
from versionhq.llm.model import DEFAULT_MODEL_NAME

"""
List up agents to be called across the project.
"""

client_manager = Agent(role="Client Manager", goal="communicate with clients on the task progress", llm=DEFAULT_MODEL_NAME)

task_evaluator = Agent(
    role="Task Evaluator",
    goal="score the output according to the given evaluation criteria.",
    llm=DEFAULT_MODEL_NAME,
    llm_config=dict(top_p=0.8, top_k=30, max_tokens=5000, temperature=0.9)
)



import os

from versionhq.agent.model import Agent
from versionhq.task.model import Task, ResponseField, TaskOutput
from versionhq.team.model import Team, TeamMember, TaskHandlingProcess, TeamOutput
from versionhq._utils.usage_metrics import UsageMetrics
from versionhq.llm.llm_vars import MODELS
from versionhq.llm.model import DEFAULT_MODEL_NAME, LLM


def test_form_team():
    agent_a = Agent(role="agent a", goal="My amazing goals", backstory="My amazing backstory", llm=DEFAULT_MODEL_NAME, max_tokens=3000)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME, max_tokens=3000)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[
            ResponseField(title="test1", data_type=str, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    task_2 = Task(
        description="Define the cohort.",
        response_fields=[
            ResponseField(title="test1", data_type=int, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=True, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
        ],
    )

    assert team.id is not None
    assert team.key is not None
    assert isinstance(team.key, str)
    assert team.managers is not None
    assert task_1 in team.manager_tasks
    assert len(team.tasks) == 2
    for item in team.tasks:
        assert item.id is task_1.id or item.id is task_2.id


def test_form_team_without_leader():
    agent_a = Agent(role="agent a", goal="My amazing goals", backstory="My amazing backstory", llm=DEFAULT_MODEL_NAME, max_tokens=3000)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME, max_tokens=3000)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[
            ResponseField(title="test1", data_type=str, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ]
    )
    task_2 = Task(
        description="Define the cohort.",
        response_fields=[
            ResponseField(title="test1", data_type=int, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
        ],
    )

    assert team.id is not None
    assert team.key is not None
    assert isinstance(team.key, str)
    assert team.managers is None
    assert team.manager_tasks is None
    assert len(team.tasks) == 2
    for item in team.tasks:
        assert item.id is task_1.id or item.id is task_2.id


def test_kickoff_without_leader():
    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[
            ResponseField(title="test1", data_type=str, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    task_2 = Task(
        description="Define the cohort.",
        response_fields=[
            ResponseField(title="test1", data_type=int, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
        ],
    )
    res = team.kickoff()
    res_all = res.return_all_task_outputs()

    assert isinstance(res, TeamOutput)
    assert res.team_id is team.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is None
    for item in res.task_output_list:
        assert isinstance(item, TaskOutput)
    assert isinstance(res_all, list)
    assert len(res_all) == 2
    for item in res_all:
        assert isinstance(item, dict)
    assert isinstance(res.token_usage, UsageMetrics)
    assert res.token_usage.total_tokens == 0 # as we dont set token usage on agent


def team_kickoff_with_task_callback():
    """
    Each task has callback with callback kwargs.
    """
    demo_list = []
    def demo_callback(item: str) -> None:
        demo_list.append(item)

    agent_a = Agent(
        role="agent a",
        goal="My amazing goals",
        llm=DEFAULT_MODEL_NAME
    )

    agent_b = Agent(
        role="agent b",
        goal="My amazing goals",
        llm=DEFAULT_MODEL_NAME
    )

    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="test1", data_type=str, required=True),],
        callback=demo_callback,
        callback_kwargs=dict(item="pytest demo 1")
    )

    task_2 = Task(
        description="Define the cohort.",
        response_fields=[ResponseField(title="test1", data_type=int, required=True),],
        callback=demo_callback,
        callback_kwargs=dict(item="pytest demo 2")
    )

    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
        ],
    )
    res = team.kickoff()

    assert res.raw is not None
    assert res.json_dict is not None
    assert len(res.return_all_task_outputs()) == 2
    assert len(demo_list) == 2
    assert "pytest" in demo_list[0]
    assert "pytest" in demo_list[1]


def test_delegate_in_team():
    """
    When the agent belongs to the team, the team manager or peers are prioritized to delegete the task.
    """

    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="test1", data_type=str, required=True),],
        allow_delegation=True
    )
    task_2 = Task(
        description="Define the cohort.",
        response_fields=[ResponseField(title="test1", data_type=int, required=True),],
        allow_delegation=False
    )
    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
        ],
    )
    res = team.kickoff()

    assert res.raw is not None
    assert res.json_dict is not None
    assert "agent b" in task_1.processed_by_agents


def test_kickoff_with_leader():
    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    task_2 = Task(
        description="Define the cohort timeframe.",
        response_fields=[
            ResponseField(title="task_2_1", data_type=int, required=True),
            ResponseField(title="task_2_2", data_type=list, required=True),
        ],
    )
    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=True, task=task_2),
        ],
    )
    res = team.kickoff()

    assert isinstance(res, TeamOutput)
    assert res.team_id is team.id
    assert res.raw is not None
    assert res.json_dict is not None
    assert team.managers[0].agent.id is agent_b.id
    assert len(res.task_output_list) == 2
    assert [item.raw is not None for item in res.task_output_list]
    assert len(team.tasks) == 2
    assert team.tasks[0].output.raw == res.raw


def test_hierarchial_process():
    """
    Manager to handle the top priority task first.
    """

    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_c = Agent(role="agent c", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    task_2 = Task(
        description="Define the cohort timeframe.",
        response_fields=[
            ResponseField(title="task_2_1", data_type=int, required=True),
            ResponseField(title="task_2_2", data_type=list, required=True),
        ],
    )
    team = Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=True, task=task_2),
            TeamMember(agent=agent_c, is_manager=False)
        ],
        process=TaskHandlingProcess.hierarchical
    )
    res = team.kickoff()

    assert isinstance(res, TeamOutput)
    assert res.team_id is team.id
    assert res.raw is not None
    assert res.json_dict is not None
    assert team.managers[0].agent.id is agent_b.id
    assert len(res.task_output_list) == 2
    assert [item.raw is not None for item in res.task_output_list]
    assert len(team.tasks) == 2
    assert team.tasks[0].output.raw == res.raw


def test_handle_team_task():
    """
    Make the best team formation with agents and tasks given.
    """

    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_c = Agent(role="agent c", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    team_task = Task(
        description="Define outbound strategies.",
        response_fields=[ResponseField(title="team_task_1", data_type=str, required=True),],
    )
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    task_2 = Task(
        description="Define the cohort timeframe.",
        response_fields=[
            ResponseField(title="task_2_1", data_type=int, required=True),
            ResponseField(title="task_2_2", data_type=list, required=True),
        ],
    )
    team_solo = Team(
        members=[
            TeamMember(agent=agent_c, is_manager=False)
        ],
        team_tasks=[team_task, task_1, task_2, ]
    )
    team_flat =  Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_c, is_manager=False)
        ],
        team_tasks=[team_task, task_2,]
    )
    team_leader =  Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=True, task=task_2),
            TeamMember(agent=agent_c, is_manager=False)
        ],
        team_tasks=[team_task, ]
    )
    team_dual_leaders =  Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=True, task=task_2),
            TeamMember(agent=agent_c, is_manager=True)
        ],
        team_tasks=[team_task, ]
    )
    team_leader_without_task =  Team(
        members=[
            TeamMember(agent=agent_a, is_manager=False, task=task_1),
            TeamMember(agent=agent_b, is_manager=False, task=task_2),
            TeamMember(agent=agent_c, is_manager=True)
        ],
        team_tasks=[team_task,]
    )
    teams = [team_solo, team_flat, team_leader, team_dual_leaders, team_leader_without_task]

    for team in teams:
        res = team.kickoff()
        assert team._get_responsible_agent(task=team_task) is not None
        assert isinstance(res, TeamOutput)
        assert res.team_id is team.id
        assert team.tasks[0].id is team_task.id
        assert res.raw is not None
        assert len(team.members) == 3
        assert len(team.tasks) == 3



import json
import threading
import datetime
import uuid
import inspect
from concurrent.futures import Future
from hashlib import md5
from typing import Any, Dict, List, Set, Optional, Tuple, Callable, Type, TypeVar
from typing_extensions import Annotated, Self

from pydantic import UUID4, BaseModel, Field, PrivateAttr, field_validator, model_validator, create_model, InstanceOf, field_validator
from pydantic_core import PydanticCustomError

from versionhq._utils.process_config import process_config
from versionhq.task import TaskOutputFormat
from versionhq.task.log_handler import TaskOutputStorageHandler
from versionhq.task.evaluate import Evaluation, EvaluationItem
from versionhq.tool.model import Tool, ToolSet
from versionhq._utils.logger import Logger


class ResponseField(BaseModel):
    """
    A class to store the response format and schema that will cascade to the LLM.
    The `config` field can store additional params:
    https://community.openai.com/t/official-documentation-for-supported-schemas-for-response-format-parameter-in-calls-to-client-beta-chats-completions-parse/932422/3
    """

    title: str = Field(default=None, description="title of the field")
    data_type: Type = Field(default=None)
    items: Optional[Type] = Field(default=None, description="store data type of the array items")
    properties: Optional[List[BaseModel]] = Field(default=None, description="store dict items in ResponseField format")
    required: bool = Field(default=True)
    nullable: bool = Field(default=False)
    config: Optional[Dict[str, Any]] = Field(default=None, description="additional rules")


    @model_validator(mode="after")
    def validate_instance(self) -> Self:
        """
        Validate the model instance based on the given `data_type`. (An array must have `items`, dict must have properties.)
        """

        if self.data_type is list and self.items is None:
            self.items = str

        if self.data_type is dict or (self.data_type is list and self.items is dict):
            if self.properties is None:
                raise PydanticCustomError("missing_properties", "The dict type has to set the properties.", {})

            else:
                for item in self.properties:
                    if not isinstance(item, ResponseField):
                        raise PydanticCustomError("invalid_properties", "Properties field must input in ResponseField format.", {})

        return self


    def _format_props(self) -> Dict[str, Any]:
        """
        Structure valid properties. We accept 2 nested objects.
        """
        from versionhq.llm.llm_vars import SchemaType

        schema_type = SchemaType(type=self.data_type).convert()
        props: Dict[str, Any] = {}

        if self.data_type is list and self.items is not dict:
            props = {
                "type": schema_type,
                "items": { "type": SchemaType(type=self.items).convert() },
            }

        elif self.data_type is list and self.items is dict:
            nested_p, nested_r = dict(), list()

            if self.properties:
                for item in self.properties:
                    nested_p.update(**item._format_props())

                    if item.required:
                        nested_r.append(item.title)

            props = {
                "type": schema_type,
                "items": {
                    "type": SchemaType(type=self.items).convert(),
                    "properties": nested_p,
                    "required": nested_r,
                    "additionalProperties": False
                }
            }

        elif self.data_type is dict:
            p, r = dict(), list()

            if self.properties:
                for item in self.properties:
                    p.update(**item._format_props())

                    # if item.required:
                    r.append(item.title)

            props = {
                "type": schema_type,
                "properties": p,
                "required": r,
                "additionalProperties": False
            }

        else:
            props = {
                "type": schema_type,
                "nullable": self.nullable,
            }

        return { self.title: { **props, **self.config }} if self.config else { self.title: props }


    def _convert(self, value: Any) -> Any:
        """
        Convert the given value to the ideal data type.
        """
        try:
            if self.type is Any:
                pass
            elif self.type is int:
                return int(value)
            elif self.type is float:
                return float(value)
            elif self.type is list or self.type is dict:
                return json.loads(eval(str(value)))
            elif self.type is str:
                return str(value)
            else:
                return value
        except:
            return value


    def create_pydantic_model(self, result: Dict, base_model: InstanceOf[BaseModel] | Any) -> Any:
        """
        Create a Pydantic model from the given result
        """
        for k, v in result.items():
            if k is not self.title:
                pass
            elif type(v) is not self.type:
                v = self._convert(v)
                setattr(base_model, k, v)
            else:
                setattr(base_model, k, v)
        return base_model


    def _annotate(self, value: Any) -> Annotated:
        """
        Address Pydantic's `create_model`
        """
        return Annotated[self.type, value] if isinstance(value, self.type) else Annotated[str, str(value)]



class TaskOutput(BaseModel):
    """
    A class to store the final output of the given task in raw (string), json_dict, and pydantic class formats.
    """

    task_id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="store Task ID")
    raw: str = Field(default="", description="Raw output of the task")
    json_dict: Dict[str, Any] = Field(default=None, description="`raw` converted to dictionary")
    pydantic: Optional[Any] = Field(default=None)
    tool_output: Optional[Any] = Field(default=None, description="store tool result when the task takes tool output as its final output")
    callback_output: Optional[Any] = Field(default=None, description="store task or agent callback outcome")
    evaluation: Optional[InstanceOf[Evaluation]] = Field(default=None, description="store overall evaluation of the task output. passed to ltm")


    def to_dict(self) -> Dict[str, Any] | None:
        """
        Convert pydantic / raw output into dict and return the dict.
        """
        return self.json_dict if self.json_dict is not None else self.pydantic.model_dump() if self.pydantic else None


    def context_prompting(self) -> str:
        """
        When the task is called as context, return its output in concise string to add it to the prompt
        """
        return json.dumps(self.json_dict) if self.json_dict else self.raw[0: 127]


    def evaluate(self, task, latency: int | float = None, tokens: int = None) -> Evaluation:
        """
        Evaluate the output based on the criteria, score each from 0 to 1 scale, and raise suggestions for future improvement.
        """
        from versionhq.task.TEMPLATES.Description import EVALUATE

        if not self.evaluation:
            self.evaluation = Evaluation()

        self.evaluation.latency = latency if latency is not None else task.latency
        self.evaluation.tokens = tokens if tokens is not None else task.tokens

        eval_criteria = task.eval_criteria if task.eval_criteria else  ["Overall competitiveness", ]

        for item in eval_criteria:
            task_1 = Task(
                description=EVALUATE.format(task_description=task.description, task_output=self.raw, eval_criteria=str(item)),
                pydantic_output=EvaluationItem
            )
            res_a = task_1.execute_sync(agent=self.evaluation.responsible_agent)
            self.evaluation.items.append(EvaluationItem(**res_a.json_dict))

        return self.evaluation


    @property
    def aggregate_score(self) -> float | int:
        if self.evaluation is None:
            return 0
        else:
            self.evaluation.aggregate_score


    @property
    def json(self) -> Optional[str]:
        if self.output_format != TaskOutputFormat.JSON:
            raise ValueError(
                """
                Invalid output format requested.
                If you would like to access the JSON output,
                pleae make sure to set the output_json property for the task
                """
            )
        return json.dumps(self.json_dict)


    def __str__(self) -> str:
        return str(self.pydantic) if self.pydantic else str(self.json_dict) if self.json_dict else self.raw



class Task(BaseModel):
    """
    Task to be executed by agents or teams.
    Each task must have a description.
    Default response is JSON string that strictly follows `response_fields` - and will be stored in TaskOuput.raw / json_dict.
    When `pydantic_output` is provided, we prioritize them and store raw (json string), json_dict, pydantic in the TaskOutput class.
    """

    __hash__ = object.__hash__
    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=True))
    _original_description: str = PrivateAttr(default=None)
    _task_output_handler = TaskOutputStorageHandler()
    config: Optional[Dict[str, Any]] = Field(default=None, description="values to set on Task class")

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="unique identifier for the object, not set by user")
    name: Optional[str] = Field(default=None)
    description: str = Field(description="Description of the actual task")

    # output
    pydantic_output: Optional[Any] = Field(default=None, description="store a custom Pydantic class as response format")
    response_fields: List[ResponseField] = Field(default_factory=list, description="store the list of ResponseFields to create the response format")
    output: Optional[TaskOutput] = Field(default=None, description="store the final task output in TaskOutput class")

    # task setup
    context: Optional[List["Task"]] = Field(default=None, description="other tasks whose outputs should be used as context")
    prompt_context: Optional[str] = Field(default=None)

    # tool usage
    tools: Optional[List[ToolSet | Tool | Any]] = Field(default_factory=list, description="tools that the agent can use aside from their tools")
    can_use_agent_tools: bool = Field(default=False, description="whether the agent can use their own tools when executing the task")
    tool_res_as_final: bool = Field(default=False, description="when set True, tools res will be stored in the `TaskOutput`")

    # execution rules
    allow_delegation: bool = Field(default=False, description="ask other agents for help and run the task instead")
    async_execution: bool = Field(default=False,description="whether the task should be executed asynchronously or not")
    callback: Optional[Callable] = Field(default=None, description="callback to be executed after the task is completed.")
    callback_kwargs: Optional[Dict[str, Any]] = Field(default_factory=dict, description="kwargs for the callback when the callback is callable")

    # evaluation
    should_evaluate: bool = Field(default=False, description="True to run the evaluation flow")
    eval_criteria: Optional[List[str]] = Field(default_factory=list, description="criteria to evaluate the outcome. i.e., fit to the brand tone")

    # recording
    processed_by_agents: Set[str] = Field(default_factory=set, description="store responsible agents' roles")
    tools_errors: int = 0
    delegations: int = 0
    latency: int | float = 0 # execution latency in sec
    tokens: int = 0 # tokens consumed


    @model_validator(mode="before")
    @classmethod
    def process_config(cls, values: Dict[str, Any]) -> None:
        return process_config(values_to_update=values, model_class=cls)


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def validate_required_fields(self) -> Self:
        required_fields = ["description",]
        for field in required_fields:
            if getattr(self, field) is None:
                raise ValueError( f"{field} must be provided either directly or through config")
        return self


    @model_validator(mode="after")
    def set_up_tools(self) -> Self:
        if not self.tools:
            pass
        else:
            tool_list = []
            for item in self.tools:
                if isinstance(item, Tool) or isinstance(item, ToolSet):
                    tool_list.append(item)
                elif (isinstance(item, dict) and "function" not in item) or isinstance(item, str):
                    pass
                else:
                    tool_list.append(item) # address custom tool
            self.tools = tool_list
        return self


    @model_validator(mode="after")
    def backup_description(self):
        if self._original_description == None:
            self._original_description = self.description
        return self


    def _draft_output_prompt(self, model_provider: str) -> str:
        """
        Draft prompts on the output format by converting `
        """

        output_prompt = ""

        if self.pydantic_output:
            output_prompt = f"""
Your response MUST STRICTLY follow the given repsonse format:
JSON schema: {str(self.pydantic_output)}
"""

        elif self.response_fields:
            output_prompt, output_formats_to_follow = "", dict()
            response_format = str(self._structure_response_format(model_provider=model_provider))
            for item in self.response_fields:
                if item:
                    output_formats_to_follow[item.title] = f"<Return your answer in {item.data_type.__name__}>"

            output_prompt = f"""
Your response MUST be a valid JSON string that strictly follows the response format. Use double quotes for all keys and string values. Do not use single quotes, trailing commas, or any other non-standard JSON syntax.
Response format: {response_format}
Ref. Output image: {output_formats_to_follow}
"""

        else:
            output_prompt = "Return your response as a valid JSON serializable string, enclosed in double quotes. Do not use single quotes, trailing commas, or other non-standard JSON syntax."

        return output_prompt


    def prompt(self, model_provider: str = None, customer: str = None, product_overview: str = None) -> str:
        """
        Format the task prompt and cascade it to the agent.
        When the task has context, add context prompting of all the tasks in the context.
        When we have cusotmer/product info, add them to the prompt.
        """

        output_prompt = self._draft_output_prompt(model_provider=model_provider)
        task_slices = [self.description, output_prompt,]

        if self.context:
            context_outputs = "\n".join([task.output.context_prompting() if hasattr(task, "output") else "" for task in self.context])
            task_slices.insert(len(task_slices), f"Consider the following context when responding: {context_outputs}")

        if self.prompt_context:
            task_slices.insert(len(task_slices), f"Consider the following context when responding: {self.prompt_context}")

        if customer:
            task_slices.insert(len(task_slices), f"Customer to address: {customer}")

        if product_overview:
            task_slices.insert(len(task_slices), f"Product to promote: {product_overview}")

        return "\n".join(task_slices)


    def _get_output_format(self) -> TaskOutputFormat:
        if self.output_json == True:
            return TaskOutputFormat.JSON
        if self.output_pydantic == True:
            return TaskOutputFormat.PYDANTIC
        return TaskOutputFormat.RAW


    def _structure_response_format(self, data_type: str = "object", model_provider: str = "gemini") -> Dict[str, Any] | None:
        """
        Structure a response format either from`response_fields` or `pydantic_output`.
        1 nested item is accepted.
        """

        from versionhq.task.structured_response import StructuredOutput

        response_format: Dict[str, Any] = None

        if self.response_fields:
            properties, required_fields = {}, []
            for i, item in enumerate(self.response_fields):
                if item:
                    if item.data_type is dict:
                        properties.update(item._format_props())
                    else:
                        properties.update(item._format_props())

                    required_fields.append(item.title)

            response_schema = {
                "type": "object",
                "properties": properties,
                "required": required_fields,
                "additionalProperties": False,
            }

            response_format = {
                "type": "json_schema",
                "json_schema": { "name": "outcome", "schema": response_schema }
            }


        elif self.pydantic_output:
            response_format = StructuredOutput(response_format=self.pydantic_output)._format()

        return response_format


    def _create_json_output(self, raw: str) -> Dict[str, Any]:
        """
        Create json (dict) output from the raw output and `response_fields` information.
        """

        if raw is None or raw == "":
            self._logger.log(level="warning", message="The model returned an empty response. Returning an empty dict.", color="yellow")
            output = { "output": "n.a." }
            return output

        try:
            r = str(raw).replace("true", "True").replace("false", "False")
            j = json.dumps(eval(r))
            output = json.loads(j)
            if isinstance(output, dict):
                return output

            else:
                r = str(raw).replace("{'", '{"').replace("{ '", '{"').replace("': '", '": "').replace("'}", '"}').replace("' }", '"}').replace("', '", '", "').replace("['", '["').replace("[ '", '[ "').replace("']", '"]').replace("' ]", '" ]').replace("{\n'", '{"').replace("{\'", '{"').replace("true", "True").replace("false", "False")
                j = json.dumps(eval(r))
                output = json.loads(j)

                if isinstance(output, dict):
                    return output

                else:
                    import ast
                    output = ast.literal_eval(r)
                    return output if isinstance(output, dict) else { "output": str(r) }

        except:
            output = { "output": str(raw) }
            return output


    def _create_pydantic_output(self, raw: str = None, json_dict: Dict[str, Any] = None) -> InstanceOf[BaseModel]:
        """
        Create pydantic output from raw or json_dict output.
        """

        output_pydantic = self.pydantic_output

        try:
            json_dict = json_dict if json_dict else self._create_json_output(raw=raw)

            for k, v in json_dict.items():
                setattr(output_pydantic, k, v)

        except:
            pass

        return output_pydantic


    def interpolate_inputs(self, inputs: Dict[str, Any]) -> None:
        """
        Interpolate inputs into the task description and expected output.
        """
        if inputs:
            self.description = self._original_description.format(**inputs)


    def _create_short_term_memory(self, agent, task_output: TaskOutput) -> None:
        """
        After the task execution, create and save short-term memory of the responsible agent.
        """

        from versionhq.agent.model import Agent
        from versionhq.memory.model import ShortTermMemory

        try:
            if isinstance(agent, Agent) and agent.use_memory == True:
                if hasattr(agent, "short_term_memory"):
                    agent.short_term_memory.save(value=task_output.raw, metadata={ "observation": self.description, }, agent=agent.role)
                else:
                    agent.short_term_memory = ShortTermMemory(agent=agent, embedder_config=agent.embedder_config)
                    agent.short_term_memory.save(value=task_output.raw, metadata={ "observation": self.description, }, agent=agent.role)

        except Exception as e:
            self._logger.log(level="error", message=f"Failed to add to short term memory: {str(e)}", color="red")
            pass


    def _create_long_term_memory(self, agent, task_output: TaskOutput) -> None:
        """
        Create and save long-term and entity memory items based on evaluation.
        """
        from versionhq.agent.model import Agent
        from versionhq.memory.model import LongTermMemory, LongTermMemoryItem

        try:
            if isinstance(agent, Agent) and agent.use_memory == True:
                evaluation = task_output.evaluation if task_output.evaluation else task_output.evaluate(task=self)

                long_term_memory_item = LongTermMemoryItem(
                    agent=str(agent.role),
                    task=str(self.description),
                    datetime=str(datetime.datetime.now()),
                    quality=evaluation.aggregate_score,
                    metadata={
                        "suggestions": evaluation.suggestion_summary,
                        "quality": evaluation.aggregate_score,
                    },
                )

                if hasattr(agent, "long_term_memory"):
                    agent.long_term_memory.save(item=long_term_memory_item)
                else:
                    agent.long_term_memory = LongTermMemory(agent=agent)
                    agent.long_term_memory.save(item=long_term_memory_item)

        except AttributeError as e:
            self._logger.log(level="error", message=f"Missing attributes for long term memory: {str(e)}", color="red")
            pass

        except Exception as e:
            self._logger.log(level="error", message=f"Failed to add to long term memory: {str(e)}", color="red")
            pass


    # task execution
    def execute_sync(self, agent, context: Optional[str] = None) -> TaskOutput:
        """
        Execute the task synchronously.
        When the task has context, make sure we have executed all the tasks in the context first.
        """

        if self.context:
            for task in self.context:
                if task.output is None:
                    task._execute_core(agent, context)

        return self._execute_core(agent, context)


    def execute_async(self, agent, context: Optional[str] = None) -> Future[TaskOutput]:
        """
        Execute the task asynchronously.
        """

        future: Future[TaskOutput] = Future()
        threading.Thread(daemon=True, target=self._execute_task_async, args=(agent, context, future)).start()
        return future


    def _execute_task_async(self, agent, context: Optional[str], future: Future[TaskOutput]) -> None:
        """
        Execute the task asynchronously with context handling.
        """

        result = self._execute_core(agent, context)
        future.set_result(result)


    def _execute_core(self, agent, context: Optional[str]) -> TaskOutput:
        """
        Execute the given task with the given agent.
        Handle 1. agent delegation, 2. tools, 3. context to consider, and 4. callbacks
        """

        from versionhq.agent.model import Agent
        from versionhq.team.model import Team

        self.prompt_context = context
        task_output: InstanceOf[TaskOutput] = None
        tool_output: str | list = None
        task_tools: List[List[InstanceOf[Tool]| InstanceOf[ToolSet] | Type[Tool]]] = []
        started_at = datetime.datetime.now()

        if self.tools:
            for item in self.tools:
                if isinstance(item, ToolSet) or isinstance(item, Tool) or type(item) == Tool:
                    task_tools.append(item)

        if self.allow_delegation:
            agent_to_delegate = None

            if hasattr(agent, "team") and isinstance(agent.team, Team):
                if agent.team.managers:
                    idling_manager_agents = [manager.agent for manager in agent.team.managers if manager.is_idling]
                    agent_to_delegate = idling_manager_agents[0] if idling_manager_agents else agent.team.managers[0]
                else:
                    peers = [member.agent for member in agent.team.members if member.is_manager == False and member.agent.id is not agent.id]
                    if len(peers) > 0:
                        agent_to_delegate = peers[0]
            else:
                agent_to_delegate = Agent(role="delegated_agent", goal=agent.goal, llm=agent.llm)

            agent = agent_to_delegate
            self.delegations += 1


        if self.tool_res_as_final == True:
            tool_output = agent.execute_task(task=self, context=context, task_tools=task_tools)
            task_output = TaskOutput(task_id=self.id, tool_output=tool_output)

        else:
            raw_output = agent.execute_task(task=self, context=context, task_tools=task_tools)
            json_dict_output = self._create_json_output(raw=raw_output)
            if "outcome" in json_dict_output:
                json_dict_output = self._create_json_output(raw=str(json_dict_output["outcome"]))

            pydantic_output = self._create_pydantic_output(raw=raw_output, json_dict=json_dict_output) if self.pydantic_output else None

            task_output = TaskOutput(
                task_id=self.id,
                raw=raw_output if raw_output is not None else "",
                pydantic=pydantic_output,
                json_dict=json_dict_output
            )

        ended_at = datetime.datetime.now()
        self.latency = (ended_at - started_at).total_seconds()

        self.output = task_output
        self.processed_by_agents.add(agent.role)

        if self.should_evaluate:
            task_output.evaluate(task=self, latency=self.latency, tokens=self.tokens)

        self._create_short_term_memory(agent=agent, task_output=task_output)
        self._create_long_term_memory(agent=agent, task_output=task_output)


        if self.callback and isinstance(self.callback, Callable):
            kwargs = { **self.callback_kwargs, **task_output.json_dict }
            sig = inspect.signature(self.callback)
            valid_keys = [param.name for param in sig.parameters.values() if param.kind == param.POSITIONAL_OR_KEYWORD]
            valid_kwargs = { k: kwargs[k] for k in valid_keys }
            callback_res = self.callback(**valid_kwargs)
            task_output.callback_output = callback_res

        # if self.output_file: ## disabled for now
        #     content = (
        #         json_output
        #         if json_output
        #         else pydantic_output.model_dump_json() if pydantic_output else result
        #     )
        #     self._save_file(content)
        return task_output


    def _store_execution_log(self, task_index: int, was_replayed: bool = False, inputs: Optional[Dict[str, Any]] = {}) -> None:
        """
        Store the task execution log.
        """

        self._task_output_handler.update(task=self, task_index=task_index, was_replayed=was_replayed, inputs=inputs)


    @property
    def key(self) -> str:
        output_format = TaskOutputFormat.JSON if self.response_fields else TaskOutputFormat.PYDANTIC if self.pydantic_output is not None else TaskOutputFormat.RAW
        source = [self.description, output_format]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()


    @property
    def summary(self) -> str:
        return f"""
Task ID: {str(self.id)}
"Description": {self.description}
"Tools": {", ".join([tool.name for tool in self.tools])}
        """



class ConditionalTask(Task):
    """
    A task that can be conditionally executed based on the output of another task.
    When the `condition` return True, execute the task, else skipped with `skipped task output`.
    """

    condition: Callable[[TaskOutput], bool] = Field(
        default=None,
        description="max. number of retries for an agent to execute a task when an error occurs",
    )


    def __init__(self, condition: Callable[[Any], bool], **kwargs):
        super().__init__(**kwargs)
        self.condition = condition
        self._logger = Logger(verbose=True)


    def should_execute(self, context: TaskOutput) -> bool:
        """
        Decide whether the conditional task should be executed based on the provided context.
        Return `True` if it should be executed.
        """
        return self.condition(context)


    def get_skipped_task_output(self):
        return TaskOutput(task_id=self.id, raw="", pydantic=None, json_dict={})


    def _handle_conditional_task(self, task_outputs: List[TaskOutput], task_index: int, was_replayed: bool) -> Optional[TaskOutput]:
        """
        When the conditional task should be skipped, return `skipped_task_output` as task_output else return None
        """

        previous_output = task_outputs[task_index - 1] if task_outputs and len(task_outputs) > 1  else None

        if previous_output and not self.should_execute(previous_output):
            self._logger.log(level="warning", message=f"Skipping conditional task: {self.description}", color="yellow")
            skipped_task_output = self.get_skipped_task_output()
            self.output = skipped_task_output

            if not was_replayed:
                self._store_execution_log(self, task_index=task_index, was_replayed=was_replayed, inputs={})
            return skipped_task_output

        return None



from typing import Dict, Any

from pydantic import BaseModel

from versionhq.agent.model import Agent
from versionhq.task.model import ResponseField
from versionhq.llm.model import DEFAULT_MODEL_NAME, LLM


class DemoChild(BaseModel):
    """
    A nested outcome class.
    """
    ch_1: str
    ch_2: dict[str, str]


class DemoOutcome(BaseModel):
    """
    A demo pydantic class to validate the outcome with various nested data types.
    """
    test0: int
    test1: float
    test2: str
    test3: bool
    test4: list[str]
    test5: dict[str, Any]
    test6: list[dict[str, Any]]
    test8: list[list[str]]
    # children: List[DemoChild]


demo_response_fields = [
    ResponseField(title="test0", data_type=int),
    ResponseField(title="test1", data_type=str, required=True),
    ResponseField(title="test2", data_type=list, items=str),
    ResponseField(title="test3", data_type=list, items=dict, properties=[
        ResponseField(title="nest1", data_type=str),
        ResponseField(title="nest2", type=dict, properties=[ResponseField(title="test", data_type=str)])
    ]),
    ResponseField(title="test4", data_type=dict, properties=[ResponseField(title="ch", data_type=tuple)]),
    ResponseField(title="test5", data_type=bool),
    ResponseField(title="test6", data_type=list, items=Any, required=False),
    # ResponseField(title="children", data_type=list, items=type(DemoChild)),
]


def create_base_agent(model: str | LLM | Dict[str, Any]) -> Agent:
    agent = Agent(role="demo", goal="My amazing goals", llm=model, max_tokens=3000, maxit=1)
    return agent


base_agent = create_base_agent(model=DEFAULT_MODEL_NAME)



import os
import uuid
import datetime
from typing import Any, Dict, List, Optional, TypeVar, Callable, Type
from typing_extensions import Self
from dotenv import load_dotenv
import litellm

from pydantic import UUID4, BaseModel, Field, InstanceOf, PrivateAttr, model_validator, field_validator, ConfigDict
from pydantic_core import PydanticCustomError

from versionhq.llm.model import LLM, DEFAULT_CONTEXT_WINDOW_SIZE, DEFAULT_MODEL_NAME
from versionhq.tool.model import Tool, ToolSet
from versionhq.knowledge.model import BaseKnowledgeSource, Knowledge
from versionhq.memory.contextual_memory import ContextualMemory
from versionhq.memory.model import ShortTermMemory, LongTermMemory, UserMemory
from versionhq._utils.logger import Logger
from versionhq.agent.rpm_controller import RPMController
from versionhq._utils.usage_metrics import UsageMetrics
from versionhq._utils.process_config import process_config


load_dotenv(override=True)
T = TypeVar("T", bound="Agent")


# def mock_agent_ops_provider():
#     def track_agent(*args, **kwargs):
#         def noop(f):
#             return f
#         return noop
#     return track_agent

# track_agent = mock_agent_ops_provider()

# agentops = None
# if os.environ.get("AGENTOPS_API_KEY"):
#     try:
#         from agentops import track_agent
#     except ImportError:
#         track_agent = mock_agent_ops_provider()
# else:
#    track_agent = mock_agent_ops_provider()


class TokenProcess:
    total_tokens: int = 0
    prompt_tokens: int = 0
    cached_prompt_tokens: int = 0
    completion_tokens: int = 0
    successful_requests: int = 0

    def sum_prompt_tokens(self, tokens: int) -> None:
        self.prompt_tokens = self.prompt_tokens + tokens
        self.total_tokens = self.total_tokens + tokens

    def sum_completion_tokens(self, tokens: int) -> None:
        self.completion_tokens = self.completion_tokens + tokens
        self.total_tokens = self.total_tokens + tokens

    def sum_cached_prompt_tokens(self, tokens: int) -> None:
        self.cached_prompt_tokens = self.cached_prompt_tokens + tokens

    def sum_successful_requests(self, requests: int) -> None:
        self.successful_requests = self.successful_requests + requests

    def get_summary(self) -> UsageMetrics:
        return UsageMetrics(
            total_tokens=self.total_tokens,
            prompt_tokens=self.prompt_tokens,
            cached_prompt_tokens=self.cached_prompt_tokens,
            completion_tokens=self.completion_tokens,
            successful_requests=self.successful_requests,
        )


# @track_agent()
class Agent(BaseModel):
    """
    Agent class that run on LLM.
    Agents execute tasks alone or in the team, using RAG tools and knowledge base if any.
    Agents will prioritize team tasks when they belong to the team.
    * (Temp) Comment out all the optional fields except for Team and LLM settings for convenience.
    """

    __hash__ = object.__hash__
    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=True))
    _rpm_controller: Optional[RPMController] = PrivateAttr(default=None)
    _request_within_rpm_limit: Any = PrivateAttr(default=None)
    _token_process: TokenProcess = PrivateAttr(default_factory=TokenProcess)
    _times_executed: int = PrivateAttr(default=0)
    config: Optional[Dict[str, Any]] = Field(default=None, exclude=True, description="values to add to the Agent class")

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    role: str = Field(description="role of the agent - used in summary and logs")
    goal: str = Field(description="concise goal of the agent (details are set in the Task instance)")
    backstory: Optional[str] = Field(default=None, description="developer prompt to the llm")
    skillsets: Optional[List[str]] = Field(default_factory=list)
    tools: Optional[List[InstanceOf[Tool | ToolSet] | Type[Tool] | Any]] = Field(default_factory=list)

    # knowledge
    knowledge_sources: Optional[List[BaseKnowledgeSource]] = Field(default=None)
    _knowledge: Optional[Knowledge] = PrivateAttr(default=None)

    # memory
    use_memory: bool = Field(default=False, description="whether to store/use memory when executing the task")
    memory_config: Optional[Dict[str, Any]] = Field(default=None, description="configuration for the memory. need to store user_id for UserMemory")
    short_term_memory: Optional[InstanceOf[ShortTermMemory]] = Field(default=None)
    long_term_memory: Optional[InstanceOf[LongTermMemory]] = Field(default=None)
    user_memory: Optional[InstanceOf[UserMemory]] = Field(default=None)
    embedder_config: Optional[Dict[str, Any]] = Field(default=None, description="embedder configuration for the agent's knowledge")

    # prompting
    use_developer_prompt: Optional[bool] = Field(default=True, description="Use developer prompt when calling the llm")
    developer_propmt_template: Optional[str] = Field(default=None, description="ddeveloper prompt template")
    user_prompt_template: Optional[str] = Field(default=None, description="user prompt template")

    # task execution rules
    team: Optional[List[Any]] = Field(default=None, description="Team to which the agent belongs")
    allow_delegation: bool = Field(default=False,description="if the agent can delegate the task to another agent or ask some help")
    max_retry_limit: int = Field(default=2 ,description="max. number of retry for the task execution when an error occurs")
    maxit: Optional[int] = Field(default=25,description="max. number of total optimization loops conducted when an error occurs")
    callbacks: Optional[List[Callable]] = Field(default_factory=list, description="callback functions to execute after any task execution")

    # llm settings cascaded to the LLM model
    llm: str | InstanceOf[LLM] | Dict[str, Any] = Field(default=None)
    function_calling_llm: str | InstanceOf[LLM] | Dict[str, Any] = Field(default=None)
    respect_context_window: bool = Field(default=True,description="Keep messages under the context window size by summarizing content")
    max_tokens: Optional[int] = Field(default=None, description="max. number of tokens for the agent's execution")
    max_execution_time: Optional[int] = Field(default=None, description="max. execution time for an agent to execute a task")
    max_rpm: Optional[int] = Field(default=None, description="max. number of requests per minute")
    llm_config: Optional[Dict[str, Any]] = Field(default=None, description="other llm config cascaded to the model")

    # cache, error, ops handling
    formatting_errors: int = Field(default=0, description="number of formatting errors.")
    agent_ops_agent_name: str = None
    agent_ops_agent_id: str = None


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="before")
    @classmethod
    def process_model_config(cls, values: Dict[str, Any]) -> None:
        return process_config(values_to_update=values, model_class=cls)


    @model_validator(mode="after")
    def validate_required_fields(self) -> Self:
        required_fields = ["role", "goal"]
        for field in required_fields:
            if getattr(self, field) is None:
                raise ValueError(f"{field} must be provided either directly or through config")
        return self


    @model_validator(mode="after")
    def set_up_llm(self) -> Self:
        """
        Set up the base model and function calling model (if any) using the LLM class.
        Pass the model config params: `llm`, `max_tokens`, `max_execution_time`, `callbacks`,`respect_context_window` to the LLM class.
        The base model is selected on the client app, else use the default model.
        """

        self.agent_ops_agent_name = self.role

        if isinstance(self.llm, LLM):
            llm = self._set_llm_params(self.llm)
            self.llm = llm

        elif isinstance(self.llm, str) or self.llm is None:
            model_name = self.llm if self.llm is not None else DEFAULT_MODEL_NAME
            llm = LLM(model=model_name)
            updated_llm = self._set_llm_params(llm)
            self.llm = updated_llm

        else:
            if isinstance(self.llm, dict):
                model_name = self.llm.pop("model_name", self.llm.pop("deployment_name", str(self.llm)))
                llm = LLM(model=model_name if model_name is not None else DEFAULT_MODEL_NAME)
                updated_llm = self._set_llm_params(llm, { k: v for k, v in self.llm.items() if v is not None })
                self.llm = updated_llm

            else:
                model_name = (getattr(self.llm, "model_name") or getattr(self.llm, "deployment_name") or str(self.llm))
                llm = LLM(model=model_name)
                llm_params = {
                    "max_tokens": (getattr(self.llm, "max_tokens") or self.max_tokens or 3000),
                    "timeout": getattr(self.llm, "timeout", self.max_execution_time),
                    "callbacks": getattr(self.llm, "callbacks", None),
                    "temperature": getattr(self.llm, "temperature", None),
                    "logprobs": getattr(self.llm, "logprobs", None),
                    "api_key": getattr(self.llm, "api_key", os.environ.get("LITELLM_API_KEY", None)),
                    "base_url": getattr(self.llm, "base_url", None),
                }
                updated_llm = self._set_llm_params(llm, llm_params)
                self.llm = updated_llm


        """
        Set up funcion_calling LLM as well.
        Check if the model supports function calling, setup LLM instance accordingly, using the same params with the LLM.
        """
        if self.function_calling_llm:
            if isinstance(self.function_calling_llm, LLM):
                if self.function_calling_llm._supports_function_calling() == False:
                    self.function_calling_llm = LLM(model=DEFAULT_MODEL_NAME)

                updated_llm = self._set_llm_params(self.function_calling_llm)
                self.function_calling_llm = updated_llm

            elif isinstance(self.function_calling_llm, str):
                llm = LLM(model=self.function_calling_llm)

                if llm._supports_function_calling() == False:
                    llm = LLM(model=DEFAULT_MODEL_NAME)

                updated_llm = self._set_llm_params(llm)
                self.function_calling_llm = updated_llm

            else:
                if isinstance(self.function_calling_llm, dict):
                    model_name = self.function_calling_llm.pop("model_name", self.function_calling_llm.pop("deployment_name", str(self.function_calling_llm)))
                    llm = LLM(model=model_name)
                    updated_llm = self._set_llm_params(llm, { k: v for k, v in self.function_calling_llm.items() if v is not None })
                    self.function_calling_llm = updated_llm

                else:
                    model_name = (getattr(self.function_calling_llm, "model_name") or getattr(self.function_calling_llm, "deployment_name") or str(self.function_calling_llm))
                    llm = LLM(model=model_name)
                    llm_params = {
                        "max_tokens": (getattr(self.function_calling_llm, "max_tokens") or self.max_tokens or 3000),
                        "timeout": getattr(self.function_calling_llm, "timeout", self.max_execution_time),
                        "callbacks": getattr(self.function_calling_llm, "callbacks", None),
                        "temperature": getattr(self.function_calling_llm, "temperature", None),
                        "logprobs": getattr(self.function_calling_llm, "logprobs", None),
                        "api_key": getattr(self.function_calling_llm, "api_key", os.environ.get("LITELLM_API_KEY", None)),
                        "base_url": getattr(self.function_calling_llm, "base_url", None),
                    }
                    updated_llm = self._set_llm_params(llm, llm_params)
                    self.function_calling_llm = updated_llm

        return self


    def _set_llm_params(self, llm: LLM, config: Dict[str, Any] = None) -> LLM:
        """
        After setting up an LLM instance, add params to the instance.
        Prioritize the agent's settings over the model's base setups.
        """

        llm.timeout = self.max_execution_time if llm.timeout is None else llm.timeout
        llm.max_tokens = self.max_tokens if self.max_tokens else llm.max_tokens

        if self.callbacks:
            llm.callbacks = self.callbacks
            llm._set_callbacks(llm.callbacks)

        if self.respect_context_window == False:
            llm.context_window_size = DEFAULT_CONTEXT_WINDOW_SIZE

        config = self.config.update(config) if self.config else config
        if config:
            valid_params = litellm.get_supported_openai_params(model=llm.model)
            for k, v in config.items():
                try:
                    if k in valid_params and v is not None:
                        setattr(llm, k, v)
                except:
                    pass
        return llm


    @model_validator(mode="after")
    def set_up_tools(self) -> Self:
        """
        Similar to the LLM set up, when the agent has tools, we will declare them using the Tool class.
        """
        if not self.tools:
            pass

        else:
            tool_list = []

            for item in self.tools:
                if isinstance(item, Tool):
                    tool_list.append(item)

                elif isinstance(item, dict) and "func" in item:
                    tool = Tool(**item)
                    tool_list.append(tool)

                elif type(item) is Tool and hasattr(item, "func"):
                    tool_list.append(item)

                else:
                    self._logger.log(level="error", message=f"Tool {str(item)} is missing a function.", color="red")
                    raise PydanticCustomError("invalid_tool", f"The tool {str(item)} is missing a function.", {})

            self.tools = tool_list

        return self


    @model_validator(mode="after")
    def set_up_backstory(self) -> Self:
        """
        Set up the backstory using a templated BACKSTORY when the backstory is None
        """

        if self.backstory is None:
            from versionhq.agent.TEMPLATES.Backstory import BACKSTORY_FULL, BACKSTORY_SHORT
            backstory = ""
            skills = ", ".join([item for item in self.skillsets]) if self.skillsets else ""
            tools = ", ".join([item.name for item in self.tools if hasattr(item, "name")]) if self.tools else ""
            role = self.role.lower()
            goal = self.goal.lower()

            if self.tools or self.skillsets:
                backstory = BACKSTORY_FULL.format(role=role, goal=goal, skills=skills, tools=tools)
            else:
                backstory = BACKSTORY_SHORT.format(role=role, goal=goal)

            self.backstory = backstory

        return self


    @model_validator(mode="after")
    def set_up_rpm(self) -> Self:
        """
        Set up RPM controller.
        """
        if self.max_rpm:
            self._rpm_controller = RPMController(max_rpm=self.max_rpm, _current_rpm=0)

        return self


    @model_validator(mode="after")
    def set_up_knowledge(self) -> Self:
        if self.knowledge_sources:
            knowledge_agent_name = f"{self.role.replace(' ', '_')}"

            if isinstance(self.knowledge_sources, list) and all(isinstance(k, BaseKnowledgeSource) for k in self.knowledge_sources):
                self._knowledge = Knowledge(
                    sources=self.knowledge_sources,
                    embedder_config=self.embedder_config,
                    collection_name=knowledge_agent_name,
                )

        return self


    @model_validator(mode="after")
    def set_up_memory(self) -> Self:
        """
        Set up memories: stm, um
        """

        if self.use_memory == True:
            self.long_term_memory = self.long_term_memory if self.long_term_memory else LongTermMemory()
            self.short_term_memory = self.short_term_memory if self.short_term_memory else ShortTermMemory(agent=self, embedder_config=self.embedder_config)

            if hasattr(self, "memory_config") and self.memory_config is not None:
                user_id = self.memory_config.get("user_id", None)
                if user_id:
                    self.user_memory = self.user_memory if self.user_memory else UserMemory(agent=self, user_id=user_id)
            else:
                self.user_memory = None

        return self


    def _train(self) -> Self:
        """
        Fine-tuned the base model using OpenAI train framework.
        """
        if not isinstance(self.llm, LLM):
            pass


    def invoke(
        self,
        prompts: str,
        response_format: Optional[Dict[str, Any]] = None,
        tools: Optional[List[InstanceOf[Tool]| InstanceOf[ToolSet] | Type[Tool]]] = None,
        tool_res_as_final: bool = False,
        task: Any = None
        ) -> Dict[str, Any]:
        """
        Create formatted prompts using the developer prompt and the agent's backstory, then call the base model.
        - Execute the task up to `self.max_retry_limit` times in case of receiving an error or empty response.
        - Pass the task_tools to the model to let them execute.
        """

        task_execution_counter = 0
        iterations = 0
        raw_response = None

        messages = []
        messages.append({ "role": "user", "content": prompts })
        if self.use_developer_prompt:
            messages.append({ "role": "system", "content": self.backstory })

        try:
            if self._rpm_controller and self.max_rpm:
                self._rpm_controller.check_or_wait()

            self._logger.log(level="info", message=f"Messages sent to the model: {messages}", color="blue")

            if tool_res_as_final:
                func_llm = self.function_calling_llm if self.function_calling_llm and self.function_calling_llm._supports_function_calling() else LLM(model=DEFAULT_MODEL_NAME)
                raw_response = func_llm.call(messages=messages, tools=tools, tool_res_as_final=True)
                task.tokens = func_llm._tokens
            else:
                raw_response = self.llm.call(messages=messages, response_format=response_format, tools=tools)
                task.tokens = self.llm._tokens

            task_execution_counter += 1
            self._logger.log(level="info", message=f"Agent response: {raw_response}", color="blue")
            return raw_response

        except Exception as e:
            self._logger.log(level="error", message=f"An error occured. The agent will retry: {str(e)}", color="red")

            while not raw_response and task_execution_counter <= self.max_retry_limit:
                while (not raw_response or raw_response == "" or raw_response is None) and iterations < self.maxit:
                    if self.max_rpm and self._rpm_controller:
                        self._rpm_controller.check_or_wait()

                    raw_response = self.llm.call(messages=messages, response_format=response_format, tools=tools)
                    task.tokens = self.llm._tokens
                    iterations += 1

                task_execution_counter += 1
                self._logger.log(level="info", message=f"Agent #{task_execution_counter} response: {raw_response}", color="blue")
                return raw_response

            if not raw_response:
                self._logger.log(level="error", message="Received None or empty response from the model", color="red")
                raise ValueError("Invalid response from LLM call - None or empty.")




    def execute_task(self, task, context: Optional[str] = None, task_tools: Optional[List[Tool | ToolSet]] = list()) -> str:
        """
        Execute the task and return the response in string.
        The agent utilizes the tools in task or their own tools if the task.can_use_agent_tools is True.
        The agent must consider the context to excute the task as well when it is given.
        """
        from versionhq.task.model import Task
        from versionhq.knowledge._utils import extract_knowledge_context

        task: InstanceOf[Task] = task
        tools: Optional[List[InstanceOf[Tool]| InstanceOf[ToolSet] | Type[Tool]]] = task_tools + self.tools if task.can_use_agent_tools else task_tools

        if self.max_rpm and self._rpm_controller:
            self._rpm_controller._reset_request_count()

        task_prompt = task.prompt(model_provider=self.llm.provider)
        if context is not task.prompt_context:
            task_prompt += context

        if self._knowledge:
            agent_knowledge = self._knowledge.query(query=[task_prompt,])
            if agent_knowledge:
                agent_knowledge_context = extract_knowledge_context(knowledge_snippets=agent_knowledge)
                if agent_knowledge_context:
                    task_prompt += agent_knowledge_context


        if self.use_memory == True:
            contextual_memory = ContextualMemory(
                memory_config=self.memory_config, stm=self.short_term_memory, ltm=self.long_term_memory, um=self.user_memory
            )
            memory = contextual_memory.build_context_for_task(task=task, context=context)
            if memory.strip() != "":
                task_prompt += memory.strip()


        # if self.team and self.team._train:
        #     task_prompt = self._training_handler(task_prompt=task_prompt)
        # else:
        #     task_prompt = self._use_trained_data(task_prompt=task_prompt)

        try:
            self._times_executed += 1
            raw_response = self.invoke(
                prompts=task_prompt,
                response_format=task._structure_response_format(model_provider=self.llm.provider),
                tools=tools,
                tool_res_as_final=task.tool_res_as_final,
                task=task
            )

        except Exception as e:
            self._times_executed += 1
            self._logger.log(level="error", message=f"The agent failed to execute the task. Error: {str(e)}", color="red")
            raw_response = self.execute_task(task, context, task_tools)

            if self._times_executed > self.max_retry_limit:
                self._logger.log(level="error", message=f"Max retry limit has exceeded.", color="red")
                raise e

        if self.max_rpm and self._rpm_controller:
            self._rpm_controller.stop_rpm_counter()

        return raw_response


    def __repr__(self):
        return f"Agent(role={self.role}, goal={self.goal}, backstory={self.backstory})"



import os
from dotenv import load_dotenv
from typing import Any, List, Optional, Dict
from pydantic import BaseModel, Field

load_dotenv(override=True)


class TeamPlanner:
    """
    Assign agents to multiple tasks.
    """

    from versionhq.task.model import Task, ResponseField, TaskOutput
    from versionhq.agent.model import Agent


    def __init__(self, tasks: List[Task], planner_llm: Optional[Any] = None):
        self.tasks = tasks
        self.planner_llm = planner_llm if planner_llm else os.environ.get("DEFAULT_MODEL_NAME")


    def _handle_assign_agents(self, unassigned_tasks: List[Task]) -> List[Any]:
        """
        Build an agent and assign it a task, then return a list of TeamMember connecting the agent created and the task given.
        """

        from versionhq.agent.model import Agent
        from versionhq.task.model import Task, ResponseField
        from versionhq.team.model import TeamMember

        new_member_list: List[TeamMember] = []
        agent_creator = Agent(
            role="agent_creator",
            goal="build an ai agent that can competitively handle the task given",
            llm=self.planner_llm,
        )

        for unassgined_task in unassigned_tasks:
            task = Task(
                description=f"""
                    Based on the following task summary, draft a AI agent's role and goal in concise manner.
                    Task summary: {unassgined_task.summary}
                """,
                response_fields=[
                    ResponseField(title="goal", data_type=str, required=True),
                    ResponseField(title="role", data_type=str, required=True),
                ],
            )
            res = task.execute_sync(agent=agent_creator)
            agent = Agent(
                role=res.json_dict["role"] if "role" in res.json_dict else res.raw,
                goal=res.json_dict["goal"] if "goal" in res.json_dict else task.description
            )
            if agent.id:
                team_member = TeamMember(agent=agent, task=unassgined_task, is_manager=False)
                new_member_list.append(team_member)

        return new_member_list



    def _handle_task_planning(self, context: Optional[str] = None, tools: Optional[str] = None) -> TaskOutput:
        """
        Handles the team planning by creating detailed step-by-step plans for each task.
        """

        from versionhq.agent.model import Agent
        from versionhq.task.model import Task

        team_planner = Agent(
            role="team planner",
            goal="Plan extremely detailed, step-by-step plan based on the tasks and tools available to each agent so that they can perform the tasks in an exemplary manner and assign a task to each agent.",
            llm=self.planner_llm,
        )

        task_summary_list = [task.summary for task in self.tasks]

        class TeamPlanIdea(BaseModel):
            plan: str | Dict[str, Any] = Field(default=None, description="a decriptive plan to be executed by the team")


        task = Task(
            description=f"""
                Based on the following task summaries, create the most descriptive plan that the team can execute most efficiently. Take all the task summaries - task's description and tools available - into consideration. Your answer only contains a dictionary.

                Task summaries: {" ".join(task_summary_list)}
            """,
            pydantic_output=TeamPlanIdea
        )
        output = task.execute_sync(agent=team_planner, context=context, tools=tools)
        return output



from pathlib import Path
from typing import Iterator, List, Optional
from urllib.parse import urlparse

try:
    from docling.datamodel.base_models import InputFormat
    from docling.document_converter import DocumentConverter
    from docling.exceptions import ConversionError
    from docling_core.transforms.chunker.hierarchical_chunker import HierarchicalChunker
    from docling_core.types.doc.document import DoclingDocument
    DOCLING_AVAILABLE = True
except ImportError:
    DOCLING_AVAILABLE = False

from pydantic import Field

from versionhq.knowledge.source import BaseKnowledgeSource
from versionhq._utils.vars import KNOWLEDGE_DIRECTORY
from versionhq._utils.logger import Logger


class DoclingSource(BaseKnowledgeSource):
    """
    Default docling class for converting documents to markdown or json.
    Support PDF, DOCX, TXT, XLSX, PPTX, MD, Images, and HTML files without any additional dependencies.
    """

    def __init__(self, *args, **kwargs):
        if not DOCLING_AVAILABLE:
            raise ImportError("The docling package is required. Please install the package using: $ uv add docling.")

        super().__init__(*args, **kwargs)

    _logger: Logger = Logger(verbose=True)
    file_paths: List[Path | str] = Field(default_factory=list)
    chunks: List[str] = Field(default_factory=list)
    safe_file_paths: List[Path | str] = Field(default_factory=list)
    content: List["DoclingDocument"] = Field(default_factory=list)
    document_converter: "DocumentConverter" = Field(
        default_factory=lambda: DocumentConverter(
            allowed_formats=[
                InputFormat.MD,
                InputFormat.ASCIIDOC,
                InputFormat.PDF,
                InputFormat.DOCX,
                InputFormat.HTML,
                InputFormat.IMAGE,
                InputFormat.XLSX,
                InputFormat.PPTX,
            ]
        )
    )

    def model_post_init(self, _) -> None:
        self.safe_file_paths = self.validate_content()
        self.content = self._load_content()


    def _load_content(self) -> List["DoclingDocument"]:
        try:
            return self._convert_source_to_docling_documents()
        except ConversionError as e:
            self._logger.log(
                level="error",
                message=f"Error loading content: {str(e)}. Supported formats: {self.document_converter.allowed_formats}",
                color="red",
            )
            raise e
        except Exception as e:
            self._logger.log(level="error", message=f"Error loading content: {e}", color="red")
            raise e


    def add(self) -> None:
        if self.content is None:
            return
        for doc in self.content:
            new_chunks_iterable = self._chunk_doc(doc)
            self.chunks.extend(list(new_chunks_iterable))
        self._save_documents()


    def _convert_source_to_docling_documents(self) -> List["DoclingDocument"]:
        conv_results_iter = self.document_converter.convert_all(self.safe_file_paths)
        return [result.document for result in conv_results_iter]


    def _chunk_doc(self, doc: "DoclingDocument") -> Iterator[str]:
        chunker = HierarchicalChunker()
        for chunk in chunker.chunk(doc):
            yield chunk.text


    def validate_content(self) -> List[Path | str]:
        processed_paths: List[Path | str] = []
        for path in self.file_paths:
            if isinstance(path, str):
                if path.startswith(("http://", "https://")):
                    try:
                        if self._validate_url(path):
                            processed_paths.append(path)
                        else:
                            raise ValueError(f"Invalid URL format: {path}")
                    except Exception as e:
                        raise ValueError(f"Invalid URL: {path}. Error: {str(e)}")
                else:
                    local_path = Path(KNOWLEDGE_DIRECTORY + "/" + path)
                    if local_path.exists():
                        processed_paths.append(local_path)
                    else:
                        raise FileNotFoundError(f"File not found: {local_path}")
            else:
                if isinstance(path, Path):
                    processed_paths.append(path)
        return processed_paths


    def _validate_url(self, url: str) -> bool:
        try:
            result = urlparse(url)
            return all(
                [
                    result.scheme in ("http", "https"),
                    result.netloc,
                    len(result.netloc.split(".")) >= 2,  # Ensure domain has TLD
                ]
            )
        except Exception:
            return False
